
#############################################################################
##
#F AutomorphismGroupPcpPGroup(R)
## 
AutomorphismGroupPcpPGroup := function( R )
    local L, M, A, a, gens, b, i, imgs, B;

    # translate 
    L := RefinedPcpGroup(R);
    M := PcpGroupToPcGroup(L); IsPGroup(M);

    # compute
    A := AutomorphismGroupPGroup(M);
    a := Concatenation(A.glAutos, A.agAutos);

    # set up
    gens := List( Igs(L), x -> PreImagesRepresentative(L!.bijection,x) );
    b := [];

    # translate back
    for i in [1..Length(a)] do
        imgs := List(Pcgs(M), x -> Image(a[i], x));
        imgs := List(imgs, x -> ExponentsOfPcElement(Pcgs(M),x));
        imgs := List(imgs, x -> MappedVector(x, Igs(L)));
        imgs := List(imgs, x -> PreImagesRepresentative(L!.bijection,x));
        b[i] := GroupHomomorphismByImagesNC(R,R,gens,imgs);
    od;

    # create group
    B := GroupByGenerators(b);
    SetSize(B, A.size);
    return B;
end;
    
#############################################################################
##
#F AutomorphismByMat( N, nat, mat )
## 
AutomorphismByMat := function( N, nat, mat )
    local t, s, tt, ss, a;

    # the action on T
    t := Igs(Source(nat));
    s := List(mat, x -> MappedVector(x,t));

    # the action on T/Tn
    tt := List( t, x -> Image( nat, x ) );
    ss := List( s, x -> Image( nat, x ) );

    # set up auto
    a := GroupHomomorphismByImagesNC( N, N, tt, ss );
    SetIsInjective(a, true);
    SetIsSurjective(a, true);
    SetIsBijective(a, true);
  
    return a;
end;

#############################################################################
##
#F CompatiblePairsInfBranch( R, mats, ser )
## 
## Compute Comp(R, T/T_n) where R = G/T acts via mats on T
##                        and ser = [T=T_0, .., T_n]
##
CompatiblePairsInfBranch := function( R, mats, ser )
    local A, mop, K, act, os, n, T, U, nat, N, flg, B, i, D, g, a, big;

    # Aut(R)
    A := AutomorphismGroupPcpPGroup(R);
    Print("Aut(R) has order ",Size(A)," \n");

    # R acts on T and T/Tn with same kernel
    mop := GroupHomomorphismByImagesNC( R, Group(mats), Igs(R), mats );
    K := Kernel(mop);

    # the action of automorphisms on subgroups
    act := function( U, a ) return Image( a, U ); end;

    # the stabilizer of K
    os := OrbitStabilizer( A, K, act );
    SetSize( os.stabilizer, Size(A)/Length(os.orbit) );
    A := os.stabilizer;
    Print("Stab(K) has order ",Size(A), " \n");

    # T/T_n and flag
    n := Length(ser);
    T := ser[1];
    U := ser[n];
    nat := NaturalHomomorphism( T, U );
    N := Image(nat);
    flg := List( ser, x -> Image(nat,x) );

    # Aut(T/T_n)
    B := AutomorphismGroupPcpPGroup(N);
    Print("Aut(T/Tn) has order ",Size(B), "\n");

    # the stabilizer of flag
    for i in [2..Length(flg)-1] do
        os := OrbitStabilizer( B, flg[i], act );
        SetSize( os.stabilizer, Size(B)/Length(os.orbit) );
        B := os.stabilizer;
    od;
    Print("Series stabilizer has order ",Size(B)," \n");
 
    # D = A x B
    D := DirectProduct(A, B);

    # the action of R on T/T_n
    g := Igs(R);
    a := List(mats, x -> AutomorphismByMat( N, nat, x ));

    # the action of D on the representations of R in Aut(N)
    big := function( a, tup )
      local new;
      new := List( g, x -> PreImagesRepresentative( tup[1], x ) );
      new := List( new, x -> MappedVector(Exponents(x), a));
      return List( new, x -> x ^ tup[2] );
    end;

    # the stabilizer
    Print("stabilize operation in group of order ",Size(D),"\n");
    os := OrbitStabilizer( D, a, big );
    SetSize( os.stabilizer, Size(D)/Length(os.orbit) );
    D := os.stabilizer;

    # these are the compatible pairs
    return D;
end;

#############################################################################
##
## ActionOfCPOnCohomology( C, cf, gens )
##
## C is a CR-record, cf a cohomology group and g a compatible pair
## Problem: C has infinite module ...
##
ActionOfCPOnCohomology := function( C, cf, gens )
    local R, n, d, p, cps, cp, g, h, m, mats, coc, H, mat, imgs, 
          tails, tail, enum, rela, i;

    # set up
    R := C.factor;
    n := Length(Igs(R));
    d := Length(C.mats[1]);
    p := Factors(Size(R))[1];

    # rewrite compatible pairs
    cps := [];
    for g in gens do
        cp := [];  
        h := Inverse(g[1]);
        cp[1] := List( Igs(R), x -> Image(h, x) );
        m := Pcp(Source(cp[2]), "snf");
        cp[2] := List( m, x -> ExponentsByPcp(m, Image(g[2],x)));
        Add( cps, cp );
    od;

    # set up for loop
    mats := [];

    # loop over basis of cohomology group cf
    for coc in cf.prei do
        H := ExtensionCR( C, coc );
        for cp in cps do

            # map imgs from R into extension H
            imgs := List( cp[1], x -> MappedVector(Exponents(x), Igs(H)) );

            # push imgs through relators
            tails := [];
            for i in [1..Length(C.enumrels)] do
                enum := C.enumrels[i];
                if enum[1] = enum[2] then
                    m := imgs[enum[1]]^p;
                elif enum[1] > enum[2] then
                    m := imgs[enum[1]]^imgs[enum[2]];
                else
                    m := imgs[enum[1]]^(imgs[enum[2]]^-1);
                fi;
                rela := C.relators[enum[1]][enum[2]];
                tail := Product(List(rela, x -> imgs[x[1]]^x[2]))^m;
                Add( tails, Exponents(tail){[n+1..n+d]} );
             od;

             # operation with g[2]
             tails := Concatenation( List( tails, x -> x * cp[2] ) );
 
             # get image in cf
             tails := SolutionIntMat(Concatenation(cf.prei, cf.denom),tails);
             tails := tails{[1..Length(cf.prei)]};
       
             # add to mat
             Add( mat, tails );
        od;
        Add( mats, mat );
    od;

    return mats;
end;

#############################################################################
##
## CohomologyInfBranch(G, n, l, t)
##
## let R_n = G/T_n and 1 <= m <= l
## compute H_{n,m} = H^2(R_n, T_n/T_{n+m})
##         A_{n,m} = Image( H^2(R_n, T_n/T_{n+m+t} ) )
##         B_{n,m} = Image( H^2(R_n, T_{n+m-t}/T_{n+m} ) )
## and the action of the compatible pairs C(R_n, T_n/T_{n+m})
##
CohomologyInfBranch := function(G, n, l, t)
    local lcs, R, act, C, cbi, cci, bas, bat, m, cfs, cc, cb, cf,
          ccA, cfA, ccB, cfB, ccI, cfI, cp, cg, cm;

    # get T_i for i <= n+l+t
    lcs := LowerCentralSeries(G, n+l+t+1);

    # get the cohomology record
    R := G/lcs[n+1];
    act := LinearActionOnPcp( Pcp(G,lcs[n+1]), Pcp(lcs[n+1]) );
    C := CRRecordByMats( R, act );

    # set up cohomology
    cbi := TwoCoboundariesCR( C );
    cci := IntTwoCocycleSystemCR( C );

    # set up lattices
    bas := [];
    bat := [];
    for m in [n+1 .. n+l+t+1] do
        bas[m] := List(Cgs(lcs[m]), x -> ExponentsByPcp(Pcp(lcs[n+1]), x));
        bat[m] := DirectSumMat( List([1..cci.len], x -> bas[m] ));
    od;

    # set up for loop
    cfs := [];

    for m in Reversed([n+2..n+l+1]) do
        cfs[m] := rec();
        Print("\n");

        # compute cohomology  H^2( R_n, T_n/T_m )
        cc := IntKernelCR( C, cci, bas[m], bat[m] );
        cb := BaseIntMat( Concatenation( cbi, bat[m] ) );
        cf := AdditiveFactorPcp(cc,cb,0);
        cfs[m].full := rec(cc := cc, cb := cb, rels := cf.rels);
        Print("H(",n,",",m,") = ", cfs[m].full.rels,"\n");

        # add big stuff A_{n,m}
        if m+t <= n+l+1 then
            ccA := BaseIntMat( Concatenation( cfs[m+t].full.cc, bat[m] ) );
            cfA := AdditiveFactorPcp(ccA, cb, 0);
            cfs[m].big := rec(cc := ccA, rels := cfA.rels);
            Print("A(",n,",",m,") = ", cfs[m].big.rels,"\n");
        fi;

        # add small stuff B_{n,m}
        if m-t >= n+1 then
            ccB := LatticeIntersection(cc, bat[m-t]);
            ccB := BaseIntMat(Concatenation(ccB, cb));
            cfB := AdditiveFactorPcp(ccB, cb, 0);
            cfs[m].sml := rec(cc := ccB, rels := cfB.rels);
            Print("B(",n,",",m,") = ", cfs[m].sml.rels,"\n");
        fi;

        # add intersection I_{n,m}
        if m+t <= n+l+1 and m-t >= n+1 then
            ccI := LatticeIntersection(cfs[m].big.cc, cfs[m].sml.cc);
            cfI := AdditiveFactorPcp(ccI, cb, 0);
            cfs[m].int := rec( cc := ccI, rels := cfI.rels);
            Print("I(",n,",",m,") = ", cfs[m].int.rels,"\n");
        fi;

        # add compatible pairs
        if m+t <= n+l+1 and m-t >= n+1 then
            cp := CompatiblePairsInfBranch( R, act, lcs{[n+1..m]});
            cg := Set(GeneratorsOfGroup(cp));
            cm := ActionOfCPOnCohomology( C, cfB, cg );
        fi;
 
    od;
end;

