TRANSLATE := true;

#############################################################################
##
#F BlowUp( l, d, ind )
##
BlowUp := function( l, d, ind )
    return Concatenation( CutVector( [1..d*l], l ){ind} );
end;

#############################################################################
##
#F EDM( mat, q )
##
EDM := function( mat, q )
    local e;
    e := ElementaryDivisorsMat(mat);
    e := List( e, x -> x mod q );
    e := Filtered(e, x -> x <> 1 );
    return Collected(e);
end;

#############################################################################
##
#F AddEquationsCC( sys, t1, t2 )
##
AddEquationsCC := function( sys, t1, t2 )
    local  i, j, v, d, t;

    t := StructuralCopy(t1);
    SubtractTailVectors( t, t2 );

    d := Length(sys.base);
    for j  in [ 1 .. sys.dim ]  do
        v := [  ];
        for i  in [ 1 .. sys.len ]  do
            if IsBound( t[i])  then
                Append( v, t[i]{[ 1 .. sys.dim ]}[j] );
            else
                Append( v, sys.zero );
            fi;
        od;
        Add( sys.base, v ); 
    od;
end;

#############################################################################
##
#F PowerWordCC( A, r, e )
##
PowerWordCC := function ( A, r, e )
    local  l;
    if Length( r ) = 1  then
        return [ [ r[1][1], e * r[1][2] ] ];
    elif e = 1  then
        return StructuralCopy(r);
    elif e > 0  then
        return Concatenation( List( [ 1 .. e ], x -> StructuralCopy(r) ) );
    elif e < 0 then
        l := InvertWord( r );
        return Concatenation( List( [ 1 .. - e ], x -> StructuralCopy(l) ) );
    fi;
end;

#############################################################################
##
#F ClearTail( A, r, e )
##
ClearTailCC := function( tail )
    local i;
    for i in [1..Length(tail)] do
        if IsBound(tail[i]) then Unbind(tail[i]); fi;
    od;
end;

#############################################################################
##
#F CollectTranslation( A, j, e, word, tail)
##
CollectTranslation := function( A, j, e, word, tail )
    local i, t, s;

    # update word
    word[j] := word[j] + e;

    # update tail
    for i in [j+1..A.gen] do
        if word[i] <> 0 then

            # get new tail
            t := [];
            s := Position( A.enumrels, [i,j] );
            t[s] := word[i] * e * A.one;

            # add it
            AddTailVectorsCR( tail, t );
        fi;
    od;
end;

#############################################################################
##
#F CollectPointGroup( A, j, word, tail, wstack, tstack, p, l )
##
CollectPointGroup := function( A, j, word, tail, wstack, tstack, p, l )
    local i, r, t, s;

    # push j through word
    for i  in [ A.gen, A.gen-1 .. j+1 ]  do
        if word[i] <> 0 then

            # update stacks
            l := l+1;
            p[l] := 1;

            # add relator to stack
            r := A.relators[i][j];
            wstack[l] := PowerWordCC( A, r, word[i] );

            # create new tail and add old one
            t := [];
            s := Position( A.enumrels, [i,j] );
            t[s] := PowerTail( A, A.relators[i][j], word[i] );
            AddTailVectorsCR( t, tail );

            # add tail to stack and clear old one
            tstack[l] := t;
            ClearTailCC( tail );
        fi;

        # reset
        word[i] := 0;
    od;

    # increase exponent
    word[j] := word[j] + 1;

    # insert power relators if exponent has reached rel order
    if word[j] = RelativeOrder( A.factor[j] ) then 

        # clear word
        word[j] := 0;

        # word is empty below g - so just insert relator
        r := StructuralCopy( A.relators[j][j] );
        for i  in [1..Length(r)] do
            word[r[i][1]] := r[i][2];
        od;

        # and just add tail
        t := []; 
        s := Position( A.enumrels, [j, j] );
        t[s] := A.one;
        AddTailVectorsCR( tail, t );
    fi;

    # return current position
    return l;
end;

#############################################################################
##
#F NormalForm1CC( A, u, v ) 
##
NormalForm1CC := function( A, u, v )
    local word, tail, wstack, tstack, l, p, i, j, e, m, r, s, t;

    # set up and push u into result
    word := StructuralCopy( u.word );
    tail := StructuralCopy( u.tail );

    # catch a trivial case
    if v.word = 0 * v.word then
        AddTailVectorsCR( tail, v.tail );
        return rec( word := word, tail := tail );
    fi;

    # create stacks and put v onto stack
    wstack := [WordOfVectorCR( v.word )];
    tstack := [StructuralCopy(v.tail)];

    # set up for loop over stacks
    l := 1;
    p := [1];                 

    # run until stacks are empty
    while l > 0 do

        # take next generator and consider cases
        j := wstack[l][p[l]][1];

        # the translation case
        if j > A.pgen then 

            # take all
            e := wstack[l][p[l]][2];
            wstack[l][p[l]][2] := 0;

            # collect
            CollectTranslation( A, j, e, word, tail );

            # check and correct stacks
            p[l] := p[l] + 1; 
            if Length(wstack[l]) < p[l]  then   
                AddTailVectorsCR( tail, tstack[l] );
                tstack[l] := 0;
                wstack[l] := [];
                l := l - 1;
            fi;
        fi;

        # the point case
        if j <= A.pgen then 

            # take one
            wstack[l][p[l]][2] := wstack[l][p[l]][2] - 1;
            if wstack[l][p[l]][2] = 0 then p[l] := p[l] + 1; fi;

            # push j through tail
            for i in [1..Length(tail)] do
                if IsBound( tail[i] ) then tail[i] := tail[i] * A.mats[j]; fi;
            od;

            # check and correct stack
            if Length(wstack[l]) < p[l]  then   
                AddTailVectorsCR( tail, tstack[l] );
                tstack[l] := 0;
                wstack[l] := [];
                l := l - 1;
            fi;

            # collect
            l := CollectPointGroup(A, j, word, tail, wstack, tstack, p, l);
        fi;
    od;

    return rec( word := word,  tail := tail );
end;

#############################################################################
##
#F CollectedTwoCC( A, u, v, q ) 
##
CollectedTwoCC := function( A, u, v, q )
    local w, i, a, b, t, s;

    w := NormalForm1CC( A, u, v );
    for i in [A.pgen+1..A.gen] do

        # get quotient and remainder
        a := RemInt(w.word[i],q);
        b := QuoInt(w.word[i],q);

        # adjust word
        w.word[i] := a;

        # adjust tail
        t := [];
        s := Length( A.enumrels ) + i - A.pgen;
        t[s] := b * A.one;
        AddTailVectorsCR ( w.tail, t );
    od;
 
    return w;
end;

#############################################################################
##
#F TwoCocyclesCC( A, q )
##
TwoCocyclesCC := function( A, q )
    local n, m, e, id, gn, pn, pa, sys1, sys2, i, j, k, w1, w2, t;

    # set up system 
    m := A.gen;
    n := A.pgen;
    e := RelativeOrdersOfPcp( A.factor );
    for i in [1..Length(e)] do if e[i] = 0 then e[i] := q; fi; od;

    sys1 := CRSystem( A.dim, Length(A.enumrels)+A.tgen, A.char );
    sys2 := CRSystem( A.dim, Length(A.enumrels)+A.tgen, A.char );

    # set up for equations 
    id := IdentityMat(m);
    gn := List( [1..m], x -> rec( word := id[x], tail := [] ) );
    pn := List( [1..m], x -> rec( word := (e[x]-1)*id[x], tail := [] ) );

    # precompute pairs gi*gj
    pa := List( [1..m], x -> [] );
    for i in [1..m]  do
        for j  in [1..i-1]  do
            pa[i][j] := NormalForm1CC( A, gn[i], gn[j] );
        od;
    od;

    # precompute powers gi^p
    for i in [1..n] do
        pa[i][i] := NormalForm1CC( A, pn[i], gn[i] ); 
    od;

    # precompute powers ti^q
    for i in [n+1..m] do
        pa[i][i] := CollectedTwoCC( A, pn[i], gn[i], q );
    od;

    # consistency 1:  k(ji) = (kj)i
    for i  in [ m, m-1 .. 1 ]  do
        for j  in [ m, m-1 .. i+1 ]  do
            for k  in [ m, m-1 .. j+1 ]  do
                w1 := CollectedTwoCC( A, gn[k], pa[j][i], q );
                w2 := CollectedTwoCC( A, pa[k][j], gn[i], q );
                if w1.word <> w2.word  then Error( "k(ji) <> (kj)i" ); fi;
                AddEquationsCC( sys1, w1.tail, w2.tail );
            od;
        od;
    od;

    # consistency 2: j^(p-1) (ji) = j^p i
    for i  in [n,n-1..1]  do
        for j  in [m,m-1..i+1]  do
            w1 := CollectedTwoCC( A, pn[j], pa[j][i], q );
            w2 := CollectedTwoCC( A, pa[j][j], gn[i], q );
            if w1.word <> w2.word  then Error( "j^(p-1) (ji) <> j^p i" ); fi;
            if j <= n then
                AddEquationsCC( sys1, w1.tail, w2.tail );
            else
                AddEquationsCC( sys2, w1.tail, w2.tail );
            fi;
        od;
    od;

    # consistency 3: j (i^p) = (ji) i^p-1
    for i  in [n,n-1..1]  do
        for j  in [m,m-1..i+1]  do
            w1 := CollectedTwoCC( A, gn[j], pa[i][i], q );
            w2 := CollectedTwoCC( A, pa[j][i], pn[i], q );
            if w1.word <> w2.word  then Error( "j i^p <> (ji) i^(p-1)" ); fi;
            AddEquationsCC( sys1, w1.tail, w2.tail );
        od;
    od;

    # consistency 4: (i^p) i = i (i^p)
    for i  in [ n, n-1 .. 1 ]  do
        w1 := CollectedTwoCC( A, pa[i][i], gn[i], q );
        w2 := CollectedTwoCC( A, gn[i], pa[i][i], q );
        if w1.word <> w2.word  then Error( "i i^p <> i^p i" ); fi;
        AddEquationsCC( sys1, w1.tail, w2.tail );
    od;

    # and return solution
    return rec( gk := TransposedMat( sys1.base ), 
                pk := TransposedMat( sys2.base ) );
end;

#############################################################################
##
#F CocycleMatrix( G, l, e )
##
CocycleMatrix := function( G, l, e )
    local U, d, p, q, T, H, iso, a, A, cm, rel, con, pou, pod, nor, 
          pow, B, C, D, M, i;

    # the translations of G
    U := SubgroupByIgs(G, Filtered(Igs(G), x -> RelativeOrder(x)=0));
    d := Length(Igs(U));

    # check
    if not l in [0..d-1] then return fail; fi;

    # set up
    p := RelativeOrder(Igs(G)[1]);
    q := p^e;

    # compute T and H 
    T := U; for i in [1..l] do T := CommutatorSubgroup(G,T); od;
    H := PcpGroupBySeries( [G, T, TrivialSubgroup(G)] );
    iso := H!.bijection; # H -> G

    # get CR record by action
    a := LinearActionOnPcp( List( Igs(H), x -> Image( iso, x ) ), Pcp(T) );
    A := CRRecordByMats( H, a );

    # add info
    A.pgen := Length(Igs(G)) - d + l;
    A.tgen := d;
    A.gen := A.pgen + A.tgen;

    # compute 2-cocyle matrix and reduce
    cm := TwoCocyclesCC( A, q );

    # split relators in A into parts
    l := Length(A.enumrels);
    con := Filtered( [1..l], x -> A.enumrels[x][2]<A.enumrels[x][1] ); 
    pou := Filtered( [1..l], x -> A.enumrels[x][1]=A.enumrels[x][2] );
    nor := BlowUp( l+A.tgen, d, Union( con, pou ) );
    pow := BlowUp( l+A.tgen, d, [l+1..l+A.tgen] );

    # set up
    A := []; B := []; C := []; D := [];

    # split matrices
    if Length(cm.gk) > 0 then 
        A := cm.gk{nor};
        C := cm.gk{pow};
    fi;
    if Length(cm.pk) > 0 then
        D := cm.pk{pow};
        B := cm.pk{nor};
    fi;

    if C <> 0*C then Print("C non-trivial\n"); fi;
    if (B mod q) <> 0*B then Print("B non-trivial\n"); fi;
#    Print("A: ",Collected(ElementaryDivisorsMat(A)),"\n");
#    Print("D: ",Collected(ElementaryDivisorsMat(D)),"\n");

    # concatenate matrices
    M := List([1..Length(A)], x -> Concatenation( A[x], B[x] ) );
    Append(M, List([1..Length(C)], x -> Concatenation( C[x], D[x] ) ) );

    return rec( A := A, B := B, C := C, D := D, M := M );
end;


