#############################################################################
##
#W classtwo.gi              Package Coclass                      Bettina Eick
##

#############################################################################
##
#F DegeneratedSubspace( mat, o ) 
##
DegeneratedSubspace := function( mat, o )
    local x, h, n;

    # catch the trivial case
    if Length(mat) = 1 then return IdentityMat(Length(mat)); fi;

    # get polynomial
    x := Indeterminate( Rationals );
    if IsPrime(o) then 
        h := x-1;
    else
        h := (x^o - 1)/CyclotomicPolynomial( Rationals, o );
    fi;
 
    # find its nullspace and return
    n := NullspaceIntMat( Value( h, mat ) );
    return TriangulizedIntegerMat( n );
end;

#############################################################################
##
#F CoverAction( T, mat ) 
##
CoverAction := function( T, mat )
    local d, e, a, A, t, r, aut, img, big;

    # set up
    d := Length(mat);
    e := Length(Igs(T));

    # find subspace
    a := Igs(T){[d+1..e]};
    A := SubgroupByIgs(T, a);

    # get automorphism
    t := Igs(T){[1..d]};
    r := List( mat, x -> MappedVector(x,t) );
    aut := GroupHomomorphismByImagesNC( T, T, t, r );

    # restrict to A
    img := List( a, x -> Image(aut,x) );
    big := List( img, x -> ExponentsByPcp(Pcp(A), x) );

    # return
    return rec( mat := big, aut := aut );
end;

#############################################################################
##
#F ClassTwoCover( s, p, l )
##
ClassTwoCover := function( s, p, l )
    local N, G, T, W, S, B, Q, mat, ext, sub, dim, aut, lat, inv, a, b, d, 
          n, m, i, w, act;

    # get the matrix action of C on T
    N := WreathFiniteNormalizer(s,p);
    G := N!.pointgroup;
    mat := MappedVector(Exponents(Product(G!.mins)), N!.mats);

    # get the class-2 nilpotent group 
    Print("get nilpotent quotient \n");
    d := p^(s-1)*(p-1);
    n := Length(Factors(Size(G)));
    T := NilpotentQuotient(FreeGroup(d),2);

    # get action on lattice l
    lat := NormalFormIntMat( (mat-mat^0)^(l-1), 2 ).normal;
    inv := lat^-1;
    act := List(G!.mats, x -> lat * x * inv );
    mat := lat * mat * inv;

    # chop 
    Print("determine subspace \n");
    ext := CoverAction(T, mat);
    sub := DegeneratedSubspace( ext.mat, p^s );
    dim := Length(ext.mat)-Length(sub);
    Print("got ",dim/d," copies of uniserial \n");

    # compute the split extension
    Print("determine split extension \n");
    aut := List( act, x -> CoverAction(T,x).aut );
    W := PcpGroupByPcp(Pcp(G));
    S := SplitExtensionByAutomorphisms( T, W, aut );
    m := Length(Igs(S));

    # embed T and B in S 
    Print("determine embedding of subspace and factor \n");
    a := Igs(S){[n+d+1..m]};
    b := List(sub, x -> MappedVector(x,a));
    B := SubgroupByIgs(S, b);
    m := m - Length(b);

    # add info
    Q := S/B;
    Q!.prime := p;
    Q!.param := s;
    Q!.latti := l-1;
    Q!.cocla := Length(Factors(Size((G))));
    Q!.point := GroupHomomorphismByImagesNC( Q, N, Igs(Q), 
                  Concatenation(Igs(G), List([n+1..Length(Igs(Q))], 
                     x -> One(N))));
    Q!.space := NaturalHomomorphism( Q, 
                  SubgroupByIgs(Q, Igs(Q){[n+d+1..Length(Igs(Q))]}));

    # return    
    return Q;
end;

#############################################################################
##
#F UniserialSubgroupsClassTwo(Q)
##
UniserialSubgroupsClassTwo := function(Q)
    local p, s, nat, uni, i, N, tmp;
    p := Q!.prime;
    s := Q!.param;
    nat := Q!.point;
    N := Range(nat);
    uni := UniserialSubgroups(N);
    for i in [1..Length(uni)] do
        tmp := PreImage(nat, Subgroup(N, uni[i].repr));
        tmp!.prime := Q!.prime;
        tmp!.param := Q!.param;
        tmp!.cocla := Length(uni[i].repr);
        uni[i] := tmp;
    od;
    return uni;
end;

#############################################################################
##
#F ClassTwoCoversCC2_Old()
##
## This was my first attempt. It does not work, but I don't know why.
##
ClassTwoCoversCC2_Old := function()
    local cov, W, D, U, V, N;

    # set up
    cov := rec();
    
    # D27, 1st lattice
    W := ClassTwoCover(2,3,1);
    D := UniserialSubgroupsClassTwo(W)[2];
    U := SubgroupByIgs(D, Igs(D){[10..Length(Igs(D))]} );
    V := SubgroupByIgs(D, Igs(D){[ 4..Length(Igs(D))]} );
    N := MaximalSubgroupsByLayer( D, Pcp(V,U), 3 )[2].repr;   
    cov.S1 := D;
    cov.N1 := N;

    # C9
    cov.C := UniserialSubgroupsClassTwo(W)[1];

    # D27, 2nd lattice
    W := ClassTwoCover(2,3,2);
    D := UniserialSubgroupsClassTwo(W)[2];
    U := SubgroupByIgs(D, Igs(D){[10..Length(Igs(D))]} );
    V := SubgroupByIgs(D, Igs(D){[ 4..Length(Igs(D))]} );
    N := MaximalSubgroupsByLayer( D, Pcp(V,U), 3 )[2].repr;   
    cov.S2 := D;
    cov.N2 := N;

    # D27, 3rd lattice
    W := ClassTwoCover(2,3,3);
    D := UniserialSubgroupsClassTwo(W)[2];
    U := SubgroupByIgs(D, Igs(D){[10..Length(Igs(D))]} );
    V := SubgroupByIgs(D, Igs(D){[ 4..Length(Igs(D))]} );
    N := MaximalSubgroupsByLayer( D, Pcp(V,U), 3 )[2].repr;   
    cov.S3 := D;
    cov.N3 := N;

    # add some info
    cov.S1!.prime := 3; cov.S1!.param := 2;
    cov.N1!.prime := 3; cov.N1!.param := 2;
    cov.S2!.prime := 3; cov.S2!.param := 2;
    cov.N2!.prime := 3; cov.N2!.param := 2;
    cov.S3!.prime := 3; cov.S3!.param := 2;
    cov.N3!.prime := 3; cov.N3!.param := 2;
    cov.C!.prime := 3; cov.C!.param := 2;

    return cov;
end;

#############################################################################
##
#F ClassTwoCoversCC2()
##
## This is the second attempt
##
ClassTwoCoversCC2 := function()
    local cov, uni, max, W, D, U, V, max1, max2, V1, V2;

    # set up
    cov := rec( );
    
    # the parent 
    W := ClassTwoCover(2,3,1);

    # uniserial subgroups
    uni := UniserialSubgroupsClassTwo(W);
    cov.S1 := uni[2];
    cov.C := uni[1];
    cov.W := W;

    # S2 and N1 - S2 is not unique!
    D := cov.S1;
    U := SubgroupByIgs(D, Igs(D){[ 4..Length(Igs(D))]} );
    V := SubgroupByIgs(D, Igs(D){[10..Length(Igs(D))]} );
    max := List( MaximalSubgroupsByLayer( D, Pcp(U,V), 3 ), x -> x.repr );
    cov.S21 := max[1];
    cov.S22 := max[4];
    cov.N1 := max[2];

    # S3 and N2
    D := cov.S21;
    U := SubgroupByIgs(D, Igs(D){[ 4..Length(Igs(D))]} );
    V := SubgroupByIgs(D, Igs(D){[10..Length(Igs(D))]} );
    max := List( MaximalSubgroupsByLayer( D, Pcp(U,V), 3 ), x -> x.repr );
    cov.S31 := max[1];
    cov.N21 := max[2];

    D := cov.S22;
    U := SubgroupByIgs(D, Igs(D){[ 4..Length(Igs(D))]} );
    V := SubgroupByIgs(D, Igs(D){[10..Length(Igs(D))]} );
    max := List( MaximalSubgroupsByLayer( D, Pcp(U,V), 3 ), x -> x.repr );
    cov.S32 := max[3];
    cov.N22 := max[1];

    # N3
    D := cov.S31;
    U := SubgroupByIgs(D, Igs(D){[ 4..Length(Igs(D))]} );
    V := SubgroupByIgs(D, Igs(D){[10..Length(Igs(D))]} );
    max := List( MaximalSubgroupsByLayer( D, Pcp(U,V), 3 ), x -> x.repr );
    cov.N3 := max[2];

    # add some info
    cov.S1!.prime := 3; cov.S1!.param := 2;
    cov.N1!.prime := 3; cov.N1!.param := 2;
    cov.S21!.prime := 3; cov.S21!.param := 2;
    cov.S22!.prime := 3; cov.S22!.param := 2;
    cov.N21!.prime := 3; cov.N21!.param := 2;
    cov.N22!.prime := 3; cov.N22!.param := 2;
    cov.S31!.prime := 3; cov.S31!.param := 2;
    cov.S32!.prime := 3; cov.S32!.param := 2;
    cov.N3!.prime := 3; cov.N3!.param := 2;
    cov.C!.prime := 3; cov.C!.param := 2;
    cov.W!.prime := 3; cov.W!.param := 2;

    return cov;
end;

##
## S1 - 6+3l
## S2 - 7+3l
## S3 - 8+3l
## 
## N1 - 8+3l
## N2 - ?
## N3 - ?
##

