
CRLGHoms := function( mats )
    local bigs, i, b, d, m, j;

    # get action on hom( wT, T )
    bigs := [];
    for i in [1..Length(mats)] do
        b := AntiSymmetricPower( mats[i] );
        b := KroneckerProduct( TransposedMat(b^-1), mats[i] );
        b := b - b^0;
        bigs[i] := b;
    od;

    # get simultaneous fixed points
    d := Length(bigs[1]);
    m := List([1..d], x -> []);
    for i in [1..d] do
        for j in [1..Length(bigs)] do
            Append(m[i], bigs[j][i]);
        od;
    od;

    # get solution space
    return Collected(Filtered(ElementaryDivisorsMat(m), x-> x<>1));
end;

CheckDim6 := function( )
    local N, W, D, uni, mat, act, lat, l, exp;

    # set up
    N := WreathFiniteNormalizer(2,3);
    W := N!.pointgroup;
    D := UniserialSubgroups(N)[2].repr;

    # the big cycle
    mat := MappedVector(Exponents(Product(W!.mins)), N!.mats);

    # the cyclic group
    Print("\n");
    Print("C9 has homs ", CRLGHoms([mat]), "\n \n");

    # lattices
    act := [];
    for l in [1..3] do
        lat := NormalFormIntMat( (mat-mat^0)^(l-1), 2 ).normal;
        act[l] := List(W!.mats, x -> lat * x * lat^-1 );
    od;

    # the wreath product
    for l in [1..3] do
        Print("W on lattice ",l," has homs ",CRLGHoms(act[l]),"\n");
    od;
    Print("\n");

    # the group of order 27
    exp := List(D, x -> ExponentsByPcp(Pcp(W),x));
    for l in [1..3] do
        mat := List(exp, x -> MappedVector(x,act[l]));
        Print("D on lattice ",l," has homs ",CRLGHoms(mat),"\n");
    od;
end;

