
#############################################################################
##
#F PrintScale
##
PrintScaleSheetClassTwo := function( Sheet, p, l, s )
    Text( Sheet, FONTS.normal, 20, 30, "|G|");
    Text( Sheet, FONTS.normal, 90, 30, "tree");
    Text( Sheet, FONTS.normal, 20, 70, 
          Concatenation( String(p), "^", String(l)));
end;

#############################################################################
##
#F StPresPcpGroup(G)
##
StPresPcpGroup := function( G )
    local p, c, H, P;
    p := Factors(Size(G))[1];
    c := ClassPGroup(G);
    H := PcpGroupToFpGroup(G);
    P := PqStandardPresentation( H : Prime:=p, ClassBound:=c);
    return ListPresentation(P);
end;

#############################################################################
##
#F IsomtypesFactors( G, sub, r )
##
IsomtypesFactors := function( G, sub, r )
    local fac, prs, i, j, c, n, p, P, F;

    # get factors and set up
    fac := List( sub, x -> G/x );
    prs := [];
    
    # get presentations
    for i in [1..Length(fac)] do
        c := ClassPGroup(fac[i]);
        n := Length(Factors(Size(fac[i])));
        p := G!.prime;
        if n-c <> r then 
            sub[i] := false;
            fac[i] := false;
            prs[i] := false;
        else
            F := PcpGroupToFpGroup(fac[i]);
            P := PqStandardPresentation( F : Prime:=p, ClassBound:=c);
            prs[i] := ListPresentation(P);
        fi;
    od;

    # discard groups
    sub := Filtered( sub, x -> x <> false );
    fac := Filtered( fac, x -> x <> false );
    prs := Filtered( prs, x -> x <> false );

    # determine isomorphisms
    for i in [1..Length(fac)] do
        if fac[i] <> false then
            j := Position(prs, prs[i]);
            if j = i then
                sub[i]!.isom := true;
            else
                sub[i]!.isom := j; 
            fi;
        fi;
    od;

    return sub;
end;

#############################################################################
##
#F DrawClassTwoBranch( G, l, reduce )
##
DrawClassTwoBranch := function( G, l, reduce )
    local p, s, d, n, t, a, T, A, U, V, W, pcp, grp, new, i, j, k,  
          cl, CR, tmp, com, g, v, title, c;

    # set up
    p := G!.prime;
    s := G!.param;
    d := p^(s-1)*(p-1);

    # get class two module
    n := Position(RelativeOrdersOfPcp(Pcp(G)), 0)-1;
    t := Igs(G){[n+1..Length(Igs(G))]};
    a := t{[d+1..Length(t)]};
    T := SubgroupByIgs(G, t);
    A := SubgroupByIgs(G, a);

    # get coclass
    c := CoclassSpaceGroupPC( G/A );

    # jump to some point in T/T' 
    U := T;
    G!.isNormal := true;
    while Index(G,U) < p^l do
        U := CommutatorSubgroup(G,U);
        U := SubgroupByIgsAndIgs( G, Igs(U), a );
    od;

    # set up graphic sheet
    title := Concatenation( "Branch for |G| = ", String(p), "^", String(n));
    Sheet := GraphicSheet( title, 600, 700 );
    PrintScaleSheetClassTwo( Sheet, p, l, 25 );

    # set up for iteration
    grp := [[U,[100, 70]]];
    Disc( Sheet, 100, 70, 3 );

    # iterate over normal subgroup in T'
    V := A;
    i := 0;
    while Length(grp) > 0 do
        i := i + 1;
        j := 1;
        Print("layer ",i," with ",Length(grp)," groups to go \n");

        W := CommutatorSubgroup(G, V);
        pcp := Pcp(V,W);
        V := W;

        new := [];
        for cl in grp do
            CR := rec( group  := cl[1],
                       super  := Pcp(G, cl[1]),
                       factor := Pcp(cl[1], GroupOfPcp(pcp)),
                       normal := pcp );
            AddFieldCR(CR); AddRelatorsCR(CR); 
            AddOperationCR(CR); AddInversesCR(CR);

            # get G-invariant complements
            tmp := InvariantComplementsCR( CR );
            Print("    determined ",Length(tmp)," complements \n");

            # reduce
            if reduce then
                com := IsomtypesFactors( G, tmp, G!.cocla );
                com := Filtered(com, x -> IsBool(x!.isom));
                Print("   reduced to ",Length(com)," isom types \n");
            else 
                com := Filtered(tmp, x -> CoclassPGroup(G/x)=G!.cocla);
                Print("   reduced to ",Length(com)," by coclass \n");
            fi;

            # draw groups and move on
            for k in [1..Length(com)] do
                repeat
                    v := PrintVertexTreeSheet(Sheet,i,j,false,cl[2]);
                    j := j + 1;
                until not IsBool(v);
                Add( new, [com[k],v] );
            od;
        od;

        # next layer
        grp := new;
    od;
end;

MaximalS := function( L )
    local sub;
    sub := MaximalSubgroupClassesByIndex(L, 3);
    sub := List( sub, Representative );
    return sub;
end;

#############################################################################
##
#F DrawClassTwoBranchNS( G, l, cc )
##
## the non-split case uses subgroups.
##
DrawClassTwoBranchNS := function( G, l, cc )
    local p, s, d, n, t, a, T, A, U, V, W, pcp, norN, maxN, prsN, norS, r,
          maxS, prsS, i, j, k, h, title, Sheet, com, sub, prs, F, v, CR, cl;

    # set up
    p := G!.prime;
    s := G!.param;
    d := p^(s-1)*(p-1);

    # get class two module
    n := Position(RelativeOrdersOfPcp(Pcp(G)), 0)-1;
    t := Igs(G){[n+1..Length(Igs(G))]};
    a := t{[d+1..Length(t)]};
    T := SubgroupByIgs(G, t);
    A := SubgroupByIgs(G, a);

    # jump to some point in T/T' 
    U := T;
    G!.isNormal := true;
    while Index(G,U) < p^l do
        U := CommutatorSubgroup(G,U);
        U := SubgroupByIgsAndIgs( G, Igs(U), a );
    od;

    # precomute maximal subgroups
    norS := [U];
    maxS := [];
    prsS := [];

    # maximal subgroups of given coclass
    Print("  compute subgroups \n");
    sub := MaximalS(G/U);
    sub := Filtered( sub, x -> CoclassPGroup(x) = cc );

    # reduce maximal subgroups
    Print("  reduce subgroups \n \n");
    prs := List( sub, x -> StPresPcpGroup(x) );
    for j in [1..Length(sub)] do
        if not IsBool( Position(prsS, prs[j]) ) then 
            sub[j] := false; 
        else
            Add( prsS, prs[j] );
            Add( maxS, sub[j] );
        fi;
    od;
    Error("choose subgroup");

    # set up graphic sheet
    title := Concatenation( "Branch for |G| = ", String(p), "^", String(n));
    Sheet := GraphicSheet( title, 600, 700 );
    PrintScaleSheetClassTwo( Sheet, p, l-1, 25 );

    # draw maximals
    for j in [1..Length(maxS)] do
        Disc( Sheet, 100 + 25*(j-1), 70, 3 );
        maxS[j] := [100 + 25*(j-1), 70];
    od;
    
    # set up for iteration
    V := A;
    i := 0;

    # iterate over normal subgroup in T'
    while Length(norS) > 0 do
        i := i + 1;
        Print("layer ",i," with ",Length(norS)," groups to go \n");

        # get next level
        W := CommutatorSubgroup(G, V);
        pcp := Pcp(V,W);
        norN := [];
        maxN := [];
        prsN := [];

        # compute groups
        for cl in norS do
            Print("  compute complements and reduce \n");

            # get next level complements
            CR := rec( group  := cl,
                       super  := Pcp(G, cl),
                       factor := Pcp(cl, GroupOfPcp(pcp)),
                       normal := pcp );
            AddFieldCR(CR); AddRelatorsCR(CR); 
            AddOperationCR(CR); AddInversesCR(CR);
            com := InvariantComplementsCR( CR );

            # reduce complements
            com := IsomtypesFactors( G, com, G!.cocla );
            com := Filtered(com, x -> IsBool(x!.isom));
            Print("  -- got ",Length(com)," complements \n");

            Print("  compute subgroups and reduce \n");

            # maximal subgroups of given coclass
            for k in [1..Length(com)] do
                sub := MaximalS(G/com[k]);
                sub := Filtered( sub, x -> CoclassPGroup(x) = cc );
                prs := List( sub, x -> StPresPcpGroup(x) );
                for j in [1..Length(sub)] do
                    h := Position( prsN, prs[j] );
                    if IsInt(h) then
                        sub[j] := false; 
                        Add( maxN[h]!.nr, Length(norN) + k );
                    else
                        sub[j]!.nr := [Length(norN)+k];
                        Add( prsN, prs[j] );
                        Add( maxN, sub[j] );
                    fi;
                od;
            od;

            # save subgroups
            Append( norN, com );
        od;

        # draw groups
        j := 1;
        for k in [1..Length(maxN)] do
            F := StPresPcpGroup(maxN[k]/Centre(maxN[k]));
            h := Position( prsS, F );
            if not IsBool(h) then
                Print("    print group with ",maxN[k]!.nr,"\n");
                repeat
                    v := PrintVertexTreeSheet(Sheet,i,j,false,maxS[h]);
                    j := j + 1;
                until not IsBool(v);
                maxN[k] := v;
            else
                maxN[k] := false;
                prsN[k] := false;
            fi;
        od;
        Print("\n");

        # next layer
        V := W;
        prsS := Filtered( prsN, x -> not IsBool(x));
        maxS := Filtered( maxN, x -> not IsBool(x));
        norS := norN;
    od;
end;

#############################################################################
##
#F DrawClassTwoBranchNR( G, l )
##
## Draw up to isomorphsm, but compute full tree.
##
DrawClassTwoBranchNR := function( G, l )
    local p, s, d, n, t, a, T, A, U, V, W, pcp, grp, new, i, j, k, h,  
          cl, CR, tmp, com, g, v, title, prs, c;

    # set up
    p := G!.prime;
    s := G!.param;
    d := p^(s-1)*(p-1);

    # get class two module
    n := Position(RelativeOrdersOfPcp(Pcp(G)), 0)-1;
    t := Igs(G){[n+1..Length(Igs(G))]};
    a := t{[d+1..Length(t)]};
    T := SubgroupByIgs(G, t);
    A := SubgroupByIgs(G, a);

    # get coclass
    c := CoclassSpaceGroupPC( G/A );

    # jump to some point in T/T' 
    U := T;
    G!.isNormal := true;
    while Index(G,U) < p^l do
        U := CommutatorSubgroup(G,U);
        U := SubgroupByIgsAndIgs( G, Igs(U), a );
    od;
    if not Index(G,U) = p^l then Error("missed index"); fi;

    # set up graphic sheet
    title := Concatenation( "Branch for |P| = ", String(p), "^", String(n));
    Sheet := GraphicSheet( title, 600, 700 );
    PrintScaleSheetClassTwo( Sheet, p, l, 25 );

    # set up for iteration
    grp := [[U,[100, 70]]];
    Disc( Sheet, 100, 70, 3 );

    # iterate over normal subgroup in T'
    V := A;
    i := 0;
    while Length(grp) > 0 do
        i := i + 1;
        j := 1;
        Print("layer ",i," with ",Length(grp)," groups to go \n");

        W := CommutatorSubgroup(G, V);
        pcp := Pcp(V,W);
        V := W;

        new := [];
        prs := [];

        for cl in grp do
            CR := rec( group  := cl[1],
                       super  := Pcp(G, cl[1]),
                       factor := Pcp(cl[1], GroupOfPcp(pcp)),
                       normal := pcp );
            AddFieldCR(CR); AddRelatorsCR(CR); 
            AddOperationCR(CR); AddInversesCR(CR);

            # get G-invariant complements
            tmp := InvariantComplementsCR( CR );
            Print("    determined ",Length(tmp)," complements \n");

            # reduce by coclass
            if i = 1 then
                tmp := Filtered(tmp, x -> CoclassPGroup(G/x)=c);
                Print("    reduced to ",Length(tmp)," groups by coclass \n");
            fi;

            # determine canonical presentation
            com := List( tmp, x -> StPresPcpGroup(G/x));

            # draw groups and set up for next layer
            for k in [1..Length(com)] do
                h := Position( prs, com[k] );
                if not IsBool(h) then 
                    Add( new, [tmp[k], new[h][2]] );
                    Add( prs, com[k] );
                else
                    repeat
                        v := PrintVertexTreeSheet(Sheet,i,j,false,cl[2]);
                        j := j + 1;
                    until not IsBool(v);
                    Add( new, [tmp[k],v] );
                    Add( prs, com[k] );
                fi;
            od;
        od;

        # next layer
        grp := new;
    od;
end;

#############################################################################
##
#F DrawClassTwoBranchBU( G, l )
##
## Bottom-Up - usually does not work, since number of complements gets
## out of hand.
##
DrawClassTwoBranchBU := function( G, l )
    local p, s, d, n, t, w, c, r, T, W, U, V, title, Sheet, com, fac, prs, 
          i, j, h, k, L, grps, v, CR, pcp;

    # set up
    p := G!.prime;
    s := G!.param;
    d := p^(s-1)*(p-1);
    n := Position(RelativeOrdersOfPcp(Pcp(G)), 0)-1;

    # get translations and wegde
    t := Igs(G){[n+1..Length(Igs(G))]};
    w := t{[d+1..Length(t)]};
    T := SubgroupByIgs(G, t);
    W := SubgroupByIgs(G, w);

    # get coclass of space group
    c := CoclassSpaceGroupPC( G/W );
    Print("consider space group of coclass ",c," \n");

    # jump to the desired point in T/T' 
    U := T;
    G!.isNormal := true;
    while Index(G,U) < p^l do
        U := CommutatorSubgroup(G,U);
        U := SubgroupByIgsAndIgs( G, Igs(U), w );
    od;

    # get corresponding module in wegde
    V := DerivedSubgroup(U);
    r := Length(Factors(Index(W,V)));
    Print("consider space group of coclass ",c," \n");

    # set up for complement computation
    pcp := Pcp(W, V);
    CR := rec( group  := U, 
               super  := Pcp(G, U),
               factor := Pcp(U, W),
               normal := pcp );
    AddFieldCR(CR); AddRelatorsCR(CR); 
    AddOperationCR(CR); AddInversesCR(CR);

    # compute invariant complements
    com := InvariantComplementsCR( CR );
    Print("    determined ",Length(com)," complements \n");

    # if there are no complements, then that's it
    if Length(com)=0 then return; fi;


    # compute factors and reduce by isomorphism (and coclass?)
    fac := List( com, x -> G/x );
    prs := [];
    for i in [1..Length(fac)] do
        
        # a temporary check
        if not CoclassPGroup(fac[i]) = c then Error("wrong cc"); fi;
   
        # get standard presentation
        L := StPresPcpGroup(fac[i]);
        if L in prs then 
            fac[i] := false;
        else
            Add(prs, L);
        fi;
    od;
    fac := Filtered(fac, x -> not IsBool(x));
    Print("    reduced to ",Length(fac)," isomorphism types \n");

    # set up to draw tree reversed
    grps := []; 
    grps[l+r] := fac;
    grps[l] := [U,[100, 70]];

    # get layers
    for i in Reversed([l+1..l+r-1]) do

        # compute factors
        fac := List(grps[i+1], x -> x/LowerCentralSeries(x)[i-c]);
        prs := [];

        # add ancestors to grps[i+1]
        for j in [1..Length(fac)] do

            # get isom type
            L := StPresPcpGroup(fac[j]);
            h := Position(prs, L);
   
            # check
            if not IsBool(h) then
                fac[j] := false;
                grps[i+1][j] := [grps[i+1][j], h];
            else
                Add(prs, L);
                grps[i+1][j] := [grps[i+1][j], Length(prs)];
            fi;
        od;

        # set up for next round
        grps[i] := Filtered(fac, x -> not IsBool(x));
    od;

    # set up sheet
    title := Concatenation( "Branch for |P| = ", String(p), "^", String(n));
    Sheet := GraphicSheet( title, 600, 700 );
    PrintScaleSheetClassTwo( Sheet, p, l, r );

    # set up groups
    Disc( Sheet, 100, 70, 3 );

    # loop over layers
    for i in [l+1..l+r] do
        j := 1;
        for h in [1..Length(grps[i])] do

            # draw grps[i][h] with connection to grps[i][h][2]
            repeat
                v := PrintVertexTreeSheet(Sheet,i-l+1,j,false,grps[i][h][2]);
                j := j + 1;
            until not IsBool(v);

            # get connection for next layer
            if i < l+r then
                for k in [1..Length(grps[i+1])] do
                    if grps[i+1][k][2] = h then
                        grps[i+1][k][2] := v;
                    fi;
                od;
            fi;
        od;
    od;
end;


