#############################################################################
##
#W tree.gi                  Package Coclass                      Bettina Eick
##

CoclassBranchInfo := function( G, s, t )
    local F, H, K, D, grp, res, nex, des, ter, cap, fun, aut, i;

    # set up p-quotient of G
    F := PcpGroupToFpGroup( G );
    H := Pq( F : Prime := G!.prime, ClassBound := s );
    K := Pq( F : Prime := G!.prime, ClassBound := s+1 );

    # set up for iteration
    grp := [H];
    res := [];
    
    # compute iterated descendants
    i := 0;
    while Length(grp) > 0 and (IsBool(t) or i < t) do
        i := i + 1;
        nex := [];
        res[i] := [[],[]];

        for D in grp do
            des := PqDescendants( D : StepSize := 1 );
            ter := Filtered( des, IsTerminal );
            cap := Filtered( des, IsCapable );

            # sort capables
            fun := function(a,b) return a>b; end;
            aut := List( cap, x -> Size(AutomorphismGroup(x)) );
            SortParallel( aut, cap, fun );

            # cut off the main line in the first step
            if i = 1 then cap := EliminateMainLine( K, cap, aut ); fi;

            # collect results
            Append( res[i][1], cap );
            Append( res[i][2], ter );
            
            # set up for next layer
            Append( nex, cap );
        od;
        Print("level ",i," yields ",List( res[i], Length ), "\n");

        # reset for next level
        grp := nex;
    od;
    return res;
end;

CoclassBranchSkeletonInfo := function( G, s )
    local F, H, K, D, grp, res, nex, des, cap, aut, i;

    # set up p-quotient of G
    F := PcpGroupToFpGroup( G );
    H := Pq( F : Prime := G!.prime, ClassBound := s );
    K := Pq( F : Prime := G!.prime, ClassBound := s+1 );

    # set up for iteration
    grp := [H];
    res := [];
    
    # compute iterated descendants
    i := 0;
    while Length(grp) > 0 do
        i := i + 1;
        nex := [];
        res[i] := [];

        for D in grp do
            cap := PqDescendants( D : StepSize := 1, AllDescendants := false );

            # cut off the main line in the first step
            if i = 1 then 
                aut := List( cap, x -> Size(AutomorphismGroup(x)) );
                cap := EliminateMainLine( K, cap, aut ); 
            fi;

            # collect results
            Add( res[i], Length(cap) );
            
            # set up for next layer
            Append( nex, cap );
        od;
        Print("level ",i," yields ",res[i], "\n");

        # reset for next level
        grp := nex;
    od;
    return res;
end;

CoclassBranchModules := function( G, s, t )
    local F, H, K, D, U, L, P, grp, nex, des, ter, cap, fun, aut, i, r;

    if s < t then Error("s should be at least t"); fi;

    # set up p-quotient of G
    F := PcpGroupToFpGroup( G );
    H := Pq( F : Prime := G!.prime, ClassBound := s );
    K := Pq( F : Prime := G!.prime, ClassBound := s+1 );
    r := Length(Factors(Size(H)));

    Print("point group has log order ",r-s+t,"\n");
    
    # compute iterated descendants
    i := 0;
    grp := [H];
    while Length(grp) > 0 do
        i := i + 1;
        nex := [];
        Print("start layer ",i," module has log order ",i+s-t,"\n");

        for D in grp do
            des := PqDescendants( D : StepSize := 1 );
            ter := Filtered( des, IsTerminal );
            cap := Filtered( des, IsCapable );

            # sort capables
            fun := function(a,b) return a>b; end;
            aut := List( cap, x -> Size(AutomorphismGroup(x)) );
            SortParallel( aut, cap, fun );

            # cut off the main line in the first step
            if i = 1 then cap := EliminateMainLine( K, cap, aut ); fi;

            # set up for next layer
            Append( nex, cap );

            # print info
            Print("  cap: ",Length(cap), "  ter: ",Length(ter), "\n");

            # investigate module - any one will do
            Append(cap, ter);
            L := cap[1];
            U := LowerCentralSeries(L)[t+1];
            if not IsAbelian(U) then 
                P := PcGroupToPcpGroup(U);
                P := PcpGroupBySeries(PCentralSeries(P));
                PrintPcpPresentation(P);
            fi;
        od;

        # reset for next level
        grp := nex;
    od;
    return cap[1];
end;

