#############################################################################
##
##  cbbound.gi                Package coclass                  Bettina Eick
##

#############################################################################
##
#F UpperBoundCobreadth(N, uni)
##
UpperBoundCobreadth := function(N, uni)
    local G, m, GG, cl, c, g, r, e, h, d, b;

    # set up
    G := SubgroupByIgs(N, uni.repr);
    m := N!.mats;

    # get classes
    GG := PcpGroupToPcGroup(G);
    cl := ConjugacyClasses(GG);
    #Print("  got ",Length(cl)," classes \n");

    # loop over classes
    b := infinity;
    for c in cl do
        g := Representative(c);
        r := Length(Factors(Size(StabilizerOfExternalSet(c))));
        if g <> g^0 then 
            e := ExponentsOfPcElement(Pcgs(GG), g);
            h := MappedVector(e, Igs(G));
            e := ExponentsByIgs(Igs(N), h);
            h := MappedVector(e, m);
            d := Determinant(h-h^0);
            if d <> 0 then 
                d := Length(Factors(d));
                b := Minimum( b, d+r );
            fi;
            #Print("    ",d," by ",r,"\n");
        fi;
    od;
    Print("     found cobreadth bound ",b,"\n");
    return b;
end;

#############################################################################
##
#F UpperBoundCobreadthRandom(N, uni)
##
UpperBoundCobreadthRandom := function(N, uni)
    local G, m, H, L, b, l, g, e, h, k, d, c, r, iso;

    # set up
    Print("    translate to pc group \n");
    G := SubgroupByIgs(N, uni.repr);
    m := N!.mats;
    H := PcpGroupToPcGroup(G);
    
    # get special pcgs for centralizer computation
    Print("    translate to special pc group \n");
    iso := IsomorphismSpecialPcGroup(H);
    L := Image(iso);

    # loop over random elements
    b := infinity;
    l := 0;
    while l < 10 do
        g := Random(H);
        if g <> g^0 then 

            # translate back into N
            e := ExponentsOfPcElement(Pcgs(H), g);
            h := MappedVector(e, Igs(G));

            # get matrix action and determinant
            e := ExponentsByIgs(Igs(N), h);
            k := MappedVector(e, m);
            d := Determinant(k-k^0);

            # add centralizer if determinant is suitable
            if d <> 0 then 
                r := Length(Factors(Size(Centralizer(L,Image(iso,g)))));
                d := Length(Factors(d));
                if b > d+r then
                    b := Minimum( b, d+r );
                else
                    l := l+1;
                fi;
                Print("    ",d," by ",r,"\n");
            fi;
        fi;
    od;
    return b;
end;

#############################################################################
##
#F UpperBoundCobreadths(s, p)
##
UpperBoundCobreadths := function( s, p )
    local N, uni, res, i, b;

    COVEREDGP := true;
    N := WreathFiniteNormalizer(s,p);
    uni := UniserialSubgroups(N, false);
    res := [];

    for i in [1..Length(uni)] do
        Print(i,"th group \n");
        b := UpperBoundCobreadth(N, uni[i]); 
        Add(res, b);
    od;
    
    return res;
end;
        
#############################################################################
##
#F ClassBreadthBounds(s, p)
##
ClassBreadthBounds := function( s, p )
    local r, N, G, C, A, uni, res, j, t, b;

    COVEREDGP := true;
    TRANSLATE := false;
    COUNTONLY := true;

    # set up
    N := WreathFiniteNormalizer(s,p);

    # get uniserial subgroups - this is a hack
    Print("  compute uniserials \n");
    #uni := UniserialSubgroups( N, false );
    uni := UniserialSubgroupsByCentre( N, false );

    # set up for cohomology
    C := WreathInfiniteCentralizer( N!.pointgroup );
    r := Sum( List( [0..s-1], x -> p^x ) );

    # add lattices
    Print("\nlattices and fixed points \n");
    A := ActionRecord( N );
    AddActions( A, N, C );

    # loop
    res := [];
    Print("  compute extensions \n");
    for j in [1..Length(uni)] do
        Print("    ",j,"th subgroup of ",Length(uni));
        Print(" with ",Length(uni[j].repr)," gens \n");

        # cohomology and coclass
        t := CohomologiesOfPointGroup( N, A, uni[j], r );
        t := Sum(t);

        # cobreadth
        b := UpperBoundCobreadthRandom(N, uni[j]); 
        #b := UpperBoundCobreadth(N, uni[j]); 

        # save results
        Add( res, [b, t] );
        uni[j] := false;
    od;
    return res;
end;

#############################################################################
##
#F CobreadthEstimateSpaceGroup(G, j, cb)
##
CobreadthEstimateSpaceGroup := function(G, j, cb)
    local T, U, H;

    # go to factor of G
    T := Translations(G);
    U := List(Igs(T), x -> x^j);
    H := G/U;

    # check cb(H) up to limit cb
    return CobreadthRandom(H, cb);
end;
    

#############################################################################
##
#F ClassBreadthPossibleExamples(s, p, opt)
##
## If opt is true, then full computation is chosen. Otherwise random
## methods and short-cuts are used.
##
ClassBreadthPossibleExamples := function( s, p, opt )
    local r, N, G, C, A, uni, res, j, t, b, i, c;

    COVEREDGP := true;
    TRANSLATE := true;
    COUNTONLY := false;

    # set up
    N := WreathFiniteNormalizer(s,p);

    # get uniserial subgroups - shortcut if not opt
    Print("  compute uniserials \n");
    if opt then
        uni := UniserialSubgroups( N, false );
    else
        uni := UniserialSubgroupsByCentre( N, false );
    fi;

    # add lattices
    Print("\nlattices and fixed points \n");
    C := WreathInfiniteCentralizer( N!.pointgroup );
    A := ActionRecord( N );
    AddActions( A, N, C );

    # loop
    res := [];
    Print("  compute extensions \n");
    for j in [1..Length(uni)] do
        Print("    ",j,"th subgroup of ",Length(uni));
        Print(" with ",Length(uni[j].repr)," gens \n");

        # cobreadth - random methods if not opt
        if opt then
            b := UpperBoundCobreadth(N, uni[j]); 
        else
            b := UpperBoundCobreadthRandom(N, uni[j]); 
        fi;

        # get extensions of coclass c with b > c+1
        t := CohomologiesOfPointGroup( N, A, uni[j], b-2 );
        t := List(Flat(t), x -> ExtensionOfPointGroup(N,A,x));
        Print("    determined ",Length(t)," groups \n");

        # check cobreadth again
        for i in [1..Length(t)] do
            Print("      check group ",i,"\n");

            r := Coclass(t[i]);
            c := CobreadthEstimateSpaceGroup( t[i], 5, r+1 );

            # store results
            if c <= r+1 then 
                t[i] := false; 
            else
                t[i]!.cbbound := b;
                t[i]!.cbestim := c;
            fi;
        od;

        # discard unpromising groups
        t := Filtered(t, x -> not IsBool(x));
        Print("    remaining ",Length(t)," groups \n \n");

        # that's it for now
        Append(res,t);
    od;

    # now reduce to non-isomorphic groups
    if p = 2 and Length(res) > 1 then
        Print("    distinguish ",Length(res)," groups \n");
        res := DistinguishGroups( res, p, 40 );
    fi;

    # that all
    return res;
end;

