#############################################################################
##
#W  spacegrps.gi                Package Coclass                  Bettina Eick
##

#############################################################################
##
#F ExpandCohomology( A, C, cc, c, l )
##
ExpandCohomology := function( A, C, cc, c, l )
    local i, d, r, t;
    d := c * cc.factor.prei;
    d := CutVector( d, Length( C.factor ) );
    d := List( C.repwrd, x -> EvaluateCocycleMG( C, d, x ) );
    d := List( d, x -> IntVecFFE(x) * A!.fixs[l] );
    return rec( cocyc := d, latnr := l, point := C.repigs );
end;

#############################################################################
##
#F CoclassSpaces( cc, A, l, r )
##
CoclassSpaces := function( cc, A, l, r )
    local base, intb, newb, spcs, facb, i, k;

    # catch the trivial case
    if Length( cc.factor.rels ) = 0 then return [[]]; fi;

    # get fixedpoint-basis
    base := AddFixedPointBasis( A, l, r );

    # get cohomology
    intb := cc.gcc;
    newb := cc.factor.gens;
    spcs := [ newb ];

    # induce spaces into Z^1/B^1
    k := Length(intb);
    for i in [2..Length(base)] do
        intb := SumIntersectionMat( intb, base[i] )[2];
        if Length(intb) < k then 
            facb := List( intb, x -> CocToFactor( cc, x ) );
            newb := SemiEchelonMat( facb ).vectors;
            k := Length(intb);
        fi;
        Add( spcs, newb );
    od;

    Add( spcs, [] );
    return Reversed( spcs );
end;

#############################################################################
##
#F ListClasses( A, C, cc, l, oc, b, n )
##
ListClasses := function( A, C, l, cs, oc, b, n )
    local al, op, i, el, j, os;

    # set up
    al := [];
    if b = 1 then
        Add(al, [rec( cocyc:=false, latnr:=l, point:=C.repigs, cclas:=n )]);
        b := 2;
    fi;

    # check if there is anything to do
    if ForAll( [b..Length(cs)], x -> Length(cs[x-1]) = Length(cs[x]) ) then
        return al;
    fi;
 
    # get operation
    op := Group(OneCohomologyOperation(C,oc), 
          IdentityMat(Length(oc.factor.rels),oc.fld)); 

    # compute orbits
    for i in [b..Length(cs)] do
        if Length( cs[i-1] ) < Length(cs[i] ) then
            el := ElementsOutside( cs[i-1], cs[i] );
            os := List( Orbits( op, el, OnRight ), x -> x[1]);
            for j in [1..Length(os)] do
                os[j] := ExpandCohomology( A, C, oc, os[j], l );
                os[j].cclas := n - i + 1;
            od;
            Append( al, os );
        fi;
    od;
    return al;
end;

#############################################################################
##
#F CountClasses( A, C, l, cs, oc, b, n )
##
CountClasses := function( A, C, l, cs, oc, b, n )
    local al, op, co, i;

    # set up
    al := List( [1..n], x -> 0 );
    if b = 1 then al[n] := 1; b := 2; fi;

    # check if there is anything to do
    if ForAll( [b..Length(cs)], x -> Length(cs[x-1]) = Length(cs[x]) ) then
        return al;
    fi;
 
    # get operation
    op := OneCohomologyOperation(C,oc); 

    # compute orbits
    co := List( [1..Length(cs)], x -> false );
    for i in [b..Length(cs)] do
        if Length( cs[i-1] ) < Length(cs[i] ) then
            if IsBool( co[i] ) then co[i] := NrOrbits( op, cs[i] ); fi;
            if IsBool( co[i-1] ) then co[i-1] := NrOrbits( op, cs[i-1] ); fi;
            al[n-i+1] := co[i] - co[i-1];
        fi;
    od;
    return al;
end;

#############################################################################
##
#F CohomologyOfPointGroup( A, C, cc, l )
##
CohomologyOfPointGroup := function( A, C, cc, l )
    local n, b, oc, dc, cs, cl, i;

    # set up
    n := Length( C.repigs );
    b := Maximum( 1, n - cc + 1 );
    AddActionMG( C, A, l );

    Print("        determine cohom \n");
    oc := OneCohom( C );
    dc := Length( oc.factor.rels );

    Print("        determine coclass spaces in dim ", dc, "\n");
    cs := CoclassSpaces( oc, A, l, Length(C.factor) );

    # just for printing
    cl := [n];
    for i in [2..Length(cs)] do
        if Length(cs[i]) > Length(cs[i-1]) then Add(cl, n-i+1); fi;
    od;
    Print("        got coclasses ", cl, "\n");

    # set up to return results
    if COUNTONLY then 
        return CountClasses( A, C, l, cs, oc, b, n );
    else
        return ListClasses( A, C, l, cs, oc, b, n );
    fi;
end;

#############################################################################
##
#F CohomologiesOfPointGroup( N, A, cl, cc )
##
CohomologiesOfPointGroup := function( N, A, cl, cc )
    local p, n, d, U, H, l, C, c, i;

    # do a check
    p := N!.prime;
    n := Length( cl.repr );
    d := p ^ (N!.param - 1);
    if n-d > cc then return []; fi;

    # find lattice limit
    U := SubgroupByIgs( N, cl.repr );
    H := SubgroupByIgs( N, cl.norm );
    l := cl.index;

    # determine cohomology record
    C := CohomologyRecordMG( N, A, U, H );
    c := [];

    # catch a very special case
    if Length( C.factor ) = 0 then 
        if cc < 1 then 
            return [];
        elif COUNTONLY then
            return [[1]];
        else
            return [[rec( cocyc := false, latnr := 1, 
                          point := C.repigs, cclas := 1)]];
        fi;
    fi;
    
    # loop over lattices 
    for i in [1..p^l] do
        Print("      ", i,"th action \n");
        c[i] := CohomologyOfPointGroup( A, C, cc, i );
    od;

    # return
    return c;
end;
