#############################################################################
##
#W  matgrp.gi                   Package Coclass                  Bettina Eick
##

#############################################################################
##
#F IsUnimodularMat( mat )
##
IsUnimodularMat := function( mat )
    if ForAny( Flat( mat ), x -> not IsInt( x ) ) then return false; fi;
    return AbsInt( Determinant( mat ) ) = 1;
end;

#############################################################################
##
#F IsUniserialMatGroup( G, p )
##
IsUniserialMatGroup := function( G, p )
    local gen, lat, new, d, g, nor, i;
    if not Set(Factors(Size(G))) = [p] then return false; fi;
    gen := GeneratorsOfGroup(G);
    d := Length(gen[1]);
    lat := IdentityMat(d);
    for i in [1..d] do
        new := [];
        for g in gen do
            Append( new, lat * (g-g^0) );
        od;
        nor := NormalFormIntMat( new, 1 ).normal;
        nor := Product( DiagonalOfMat( nor ) );
        if not nor = p^i then return false; fi;
        lat := new;
    od;
    return true;
end;

#############################################################################
##
#F FindConjugatingMats( gens, imgs )
##
FindConjugatingMats := function( gens, imgs )
    local l, d, mats, i, mat, rhs, lhs, syst, j, null, nmat;

    l := Length(gens);
    d := Length(gens[1]);

    mats := [];
    for i in [1..l] do
        mat := IdentityMat( d );
        rhs := KroneckerProduct( TransposedMat( gens[i] ), mat );
        lhs := KroneckerProduct( mat, imgs[i] );
        mat := rhs - lhs;
        Add( mats, mat );
    od;

    syst := List( [1..d^2], x -> [] );
    for i in [1..l] do
        for j in [1..d^2] do
            Append( syst[j], mats[i][j] );
        od;
    od;

    null := NullspaceMat( syst );
    nmat := List( null, x -> MatByVector( x, d ) );
    return nmat;
end;

#############################################################################
##
#F PRootDiagonal( s, p )
##
PRootDiagonal := function( s, p )
    local x, m;
    x := Indeterminate(Rationals);
    m := CompanionMat( Sum( List( [1..p], i -> x^(i-1) ) ) );
    m := DiagonalBlockMat( List( [1..p^(s-1)], x -> m ) );
    return m;
end;

