#############################################################################
##
#W  pcpmat.gi                   Package Coclass                  Bettina Eick
##

#############################################################################
##
#F PcpToMat( G, mats, g )
##
PcpToMat := function( G, mats, g )
    if IsGroup(G) then
        return MyMappedVector( Exponents(g), mats );
    else
        return MyMappedVector( ExponentsByPcp(G, g), mats );
    fi;
end;

#############################################################################
##
#F IsTrivialBlock( mat, b, j )
##
IsTrivialBlock := function( mat, b, j )
    local c, i, k;
    c := (j-1)*b;
    for i in [1..b] do
        for k in [1..b] do
            if mat[i][k+c] <> 0 then return false; fi;
        od;
    od;
    return true;
end;

#############################################################################
##
#F ExponentsByWreathProduct( mat, s, p )
##
ExponentsByWreathProduct := function( mat, s, p )
    local j, d, b, e, i, k, blm, f;

    # the trivial case
    if s = 1 then
        j := Position( mat[1], -1 );
        if IsBool( j ) then
            return [0]; 
        else
            return [p-j];
        fi;
    fi;
    
    # dimension and block size
    d := p^(s-1) * (p-1);
    b := d / p;
    
    # find the permutation action
    j := 1;
    while j <= p and IsTrivialBlock( mat, b, j ) do
        j := j + 1;
    od;
    e := [j-1];

    # consider the individual blocks by recursion
    for i in [1..p] do
        k := p - j + i + 1; if k > p then k := k - p; fi;
        blm := mat{[(k-1)*b+1 .. k*b]}{[(i-1)*b+1 .. i*b]};
        f := ExponentsByWreathProduct( blm, s-1, p );
        Append( e, f );
    od;

    return e;
end;
    
#############################################################################
##  
#F MatToPcp( G, m )
##
MatToPcp := function( G, m )
    local e;
    e := ExponentsByWreathProduct( m, G!.param, G!.prime );
    if G!.prime = 2 then
        return MyMappedVector( e, Igs(G) );
    else
        return MyMappedVector( e, Igs(G){[G!.param+1..Length(Igs(G))]} );
    fi;
end;

#############################################################################
##
#F PcpSubgroupByMatSubgroup( G, L )
##
PcpSubgroupByMatSubgroup := function( G, L )
    local s, p, mats, e, g, U;
    s := G!.param;
    p := G!.prime;
    mats := GeneratorsOfGroup(L);
    e := List( mats, x -> ExponentsByWreathProduct( x, s, p ) );
    g := List( e, x -> MappedVector( x, Igs(G) ) );
    return Subgroup( G, g );
end;

#############################################################################
##
#F AsSubgroupOfMaximalPointGroup( G, U )
##
AsSubgroupOfMaximalPointGroup := function( G, U )

    # create a group if U is a list
    if IsList(U) and U = [] then
        U := Subgroup( G, U );
    elif IsList( U ) and IsMatrix( U[1] ) then
        U := Group( U );
    elif IsList( U ) and U[1] in G then
        U := Subgroup( G, U );
    fi;

    # adjust the group
    if IsMatrixGroup( U ) then
        U := PcpSubgroupByMatSubgroup( G, U );
    fi;
    return U;
end;




