#############################################################################
##
#W pointcent.gi                Package Coclass                   Bettina Eick
##

#############################################################################
##
#F QRootDiagonal( s, i, p )
##
QRootDiagonal := function( s, i, p )
    local f, m, d;
    f := CyclotomicPolynomial( Rationals, p^(s-i) );
    m := CompanionMat( f );
    d := DiagonalBlockMat( List( [1..p^i], x -> m ) );
    return d;
end;

#############################################################################
##
#F WreathInfiniteCentralizer( G )
##
InstallMethod( WreathInfiniteCentralizer, true, [IsGroup], 0,
function( G )
    local C, c, N, m, I, d, j, p, l, cent, i, M;

    if G!.prime = 2 then 
        M := AbelianPcpGroup( 0, [] );
        M!.prime := G!.prime;
        M!.param := G!.param;
        M!.index := G!.index;
        M!.mats := [];
        return M;
    fi;

    # get centre
    C := Centre(G);
    c := Igs(C)[1];
   
    # get matrix translation
    N := Parent(Parent(G));
    m := MappedVector(Exponents(c), N!.mats);
    I := m^0;
    d := (m - I)^-1;

    # setup
    j := G!.param - G!.index;
    p := G!.prime;
    l := Filtered([3..p^j], x -> (Gcd(x,p) = 1 and not IsInt(x/2)) );
    
    # construct matrices
    cent := [];
    for i in l do
        Add( cent, (m^i - I) * d );
    od;

    # do a check
    if not ForAll(cent, x -> Determinant(x) = 1 ) then 
        Error("wrong cent");
    fi;

    # construct pcp group
    M := AbelianPcpGroup( Length(l), List(l, x -> 0) );
    M!.prime := G!.prime;
    M!.param := G!.param;
    M!.index := G!.index;
    M!.mats := cent;

    return M;
end );


