#############################################################################
##
#W spacegrp.gi                 Package Coclass                   Bettina Eick
##

#############################################################################
##
#F ExtensionsOfPointGroup( G, cl )
##
ExtensionsOfPointGroup := function( G, cl )
    local new;
    new := Flat( CohomologiesOfPointGroup( G, cl ) );
    return List( new, x -> ExtensionOfPointGroup(G,x) );
end;

#############################################################################
##
#F ExtensionsOfPointGroups( G, uni, cc, start )
##
ExtensionsOfPointGroups := function( G, uni, cc, start )
    local new, i;
    new := [];
    for i in [start..Length(uni)] do
        Print("  ",i,"th subgroup \n");
        new[i] := CohomologiesOfPointGroup( G, uni[i], cc );
        uni[i] := false;
    od;
end;

#############################################################################
##
#F ClassifySpaceGroups( s, p, cc ) 
##
InstallMethod( ClassifySpaceGroups, true, [IsInt, IsInt, IsInt], 0,
function( s, p, cc )
    local N, A, res, i, Ni, Gi, Ci, uni, len, j, t;

    # check 
    if ForAny( [s,p,cc], x -> not IsInt(x)) then Error("wrong input"); fi;
    if not IsPrimeInt(p) then Error("wrong input"); fi;

    # get cc into line
    cc := Minimum(cc, Sum( List( [0..s-1], x -> p^x ) ) );
    if cc < s then return []; fi;

    # get the maximal point group and the relevant actions
    Print("\npoint group \n");
    N := WreathFiniteNormalizer( s, p );

    # add lattices
    Print("\nlattices and fixed points \n");
    A := ActionRecord( N );

    # loop over centres
    res := [];
    for i in [0..s-1] do
 
        Print("\nset up for groups with centre ",i,"\n");
        Ni := WreathFiniteNormalizer( N, i );
        Gi := Ni!.pointgroup;
        Ci := WreathInfiniteCentralizer( Gi );

        Print("  add actions \n");
        AddActions( A, Ni, Ci );

        # get uniserial subgroups 
        Print("  compute uniserials \n");
        uni := UniserialSubgroupsByCentre( Ni, cc );
        len := List( uni, x -> Length(x.repr) );
        SortParallel( len, uni );

        # loop
        Print("  compute extensions \n");
        for j in [1..Length(uni)] do
            if not IsBool(uni[j]) then 
                Print("    ",j,"th subgroup of ",Length(uni));
                Print(" with ",Length(uni[j].repr)," gens \n");
                if SPLITONLY then 
                    t := List([1..p^uni[j].index], x -> 
                         [rec(cocyc := false, 
                             latnr := x, 
                             point := uni[j].repr,
                             cclas := Length(uni[j].repr))]);
                else
                    t := CohomologiesOfPointGroup( N, A, uni[j], cc );
                fi;
                Add( res, t );

                if COUNTONLY and Sum(t) = 0 * t[1] then 
                    DiscardSupergroups( N, uni, uni[j] );
                elif not COUNTONLY and Length(Flat(t)) = 0 then
                    DiscardSupergroups( N, uni, uni[j] );
                fi;
            fi;
        od;
    od;
    Print("\n");

    # set up results
    if COUNTONLY then
        res := Concatenation(res);
        return Sum(res);
    elif TRANSCODE then
        res := Flat(res);
        for i in [1..Length(res)] do
            res[i] := ExtensionOfPointGroup(N,A,res[i]);
            if res[i]!.cclas = cc then
                res[i] := CodePcpGroup(res[i]);
            else
               res[i] := false;
            fi;
        od;
        return Filtered( res, x -> not IsBool(x));
    elif TRANSLATE then
        res := Flat(res);
        return List( res, x -> ExtensionOfPointGroup(N,A,x) );
    else
        res := Concatenation(res);
        return rec( group := N, exts := res );
    fi;
end );

InstallOtherMethod( ClassifySpaceGroups, true, [IsInt, IsInt], 0,
function(s, p)
    local cc;
    cc := Sum( List( [0..s-1], x -> p^x ) );
    return ClassifySpaceGroups(s, p, cc);
end );

#############################################################################
##
#F FilterByGroup( N, todo, uni, U )
##
FilterByGroup := function( N, todo, uni, U )
    local i, j, H;
    for i in [1..Length(todo)] do
        j := todo[i];
        if uni[j].repr = U.repr then todo[i] := false; fi;
        if Length(uni[j].repr) > Length(U.repr) then 
            H := SubgroupByIgs(N, uni[j].repr);
            if ForAll( U.repr, x -> x in H ) then todo[i] := false; fi;
        fi;
    od;
    return Filtered(todo, x -> not IsBool(x));
end;

#############################################################################
##
#F ClassifySpaceGroupsByUnis( N, uni, cc ) 
##
ClassifySpaceGroupsByUnis := function( N, uni, cc )
    local p, s, A, res, todo, sudo, lens, i, Ni, Gi, Ci, j, t;

    p := N!.prime;
    s := N!.param;
    #COUNTONLY := true;

    # get cc into line
    cc := Minimum(cc, Sum( List( [0..s-1], x -> p^x ) ) );
    if cc < s then return []; fi;

    # add lattices
    Print("\nlattices and fixed points \n");
    A := ActionRecord( N );

    # set up for iteration
    res := [];
    todo := [1..Length(uni)];
    lens := List( uni, x -> Length(x.repr) );
    SortParallel( lens, todo );

    # loop over centres
    for i in [0..s-1] do
 
        Print("\nset up for groups with centre ",i,"\n");
        Ni := WreathFiniteNormalizer( N, i );
        Gi := Ni!.pointgroup;
        Ci := WreathInfiniteCentralizer( Gi );

        Print("  add actions \n");
        AddActions( A, Ni, Ci );

        # loop
        Print("  compute extensions \n");
        sudo := Filtered(todo, x -> uni[x].index = i );
        while Length(sudo) > 0 do
            j := sudo[1];
            Print("    next of ",Length(sudo)," with ");
            Print(Length(uni[j].repr)," gens \n");
            t := CohomologiesOfPointGroup( N, A, uni[j], cc );
            t := Sum(t);
            if t = 0*t then
                todo := FilterByGroup( N, todo, uni, uni[j] );
                sudo := Filtered( sudo, x -> x in todo );
            else
                Add( res, [j,t] );
                sudo := sudo{[2..Length(sudo)]};
                todo := Filtered( todo, x -> x <> j );
            fi;
            Print("    ", Length(sudo)," in ", Length(todo)," remain\n");
        od;
    od;
    Print("\n");

    # set up results
    return res;
end;

#############################################################################
##
#F PointGroupsWithExtension( s, p, cc, cent ) 
##
PointGroupsWithExtension := function( s, p, cc, cent )
    local N, A, U, Ni, Gi, Ci, res, uni, len, i, j, t;

    # get cc into line
    cc := Minimum(cc, Sum( List( [0..s-1], x -> p^x ) ) );
    if cc < s then return []; fi;

    # get cent into line
    if IsBool( cent ) then cent := [0..s-1]; fi;

    # set flags
    COUNTONLY := true;
    TRANSLATE := false;
    COVEREDGP := false;

    # get the point group normalizer
    Print("\npoint group \n");
    N := WreathFiniteNormalizer( s, p );

    # add lattices
    Print("lattices and fixed points \n");
    A := ActionRecord( N );

    # loop over centres
    res := [];
    for i in cent do
 
        Print("set up for groups with centre ",i,"\n");
        Ni := WreathFiniteNormalizer( N, i );
        Gi := Ni!.pointgroup;
        Ci := WreathInfiniteCentralizer( Gi );

        Print("  add actions \n");
        AddActions( A, Ni, Ci );

        # get uniserial subgroups 
        Print("  compute uniserials \n");
        uni := UniserialSubgroupsByCentre( Ni, cc );
        len := List( uni, x -> Length(x.repr) );
        SortParallel( len, uni );

        # loop
        Print("  try extensions \n");
        for j in [1..Length(uni)] do
            if not IsBool(uni[j]) then 
                Print("    ",j,"th subgroup of ",Length(uni));
                Print(" with ",Length(uni[j].repr)," gens \n");
                t := CohomologiesOfPointGroup( N, A, uni[j], cc );

                # evaluate results
                if Sum(t) = 0 * t[1] then 
                    DiscardSupergroups( N, uni, uni[j] );
                else
                    U := SubgroupByIgs(N, uni[j].repr);
                    U!.unirc := uni[j];
                    U!.cohom := t;
                    Add( res, U );
                fi;
            fi;
        od;
    od;
    Print("\n");

    # set up results
    return res;
end;


