#############################################################################
##
#W library.gi                    Coclass Package                 Bettina Eick
##

#############################################################################
##
#A DiagonalBlockMat(blocks)
##
DiagonalBlockMat := function( bl )
    local d, M, s, i, t;
    d := Sum( List(bl, Length) );
    M := MutableNullMat(d,d);
    s := 0;
    for i in [1..Length(bl)] do
        t := Length(bl[i]);
        M{[s+1..s+t]}{[s+1..s+t]} := bl[i];
        s := s+t;
    od;
    M := M*bl[1][1][1]^0;
    return M;
end;

#############################################################################
##
#A CodePcpGroup(G)
##
CodePcpGroup := function(G)
    local g, r, n, i, j, w, rels;

    g := Pcp(G);
    r := RelativeOrdersOfPcp(g);
    n := Length(g);

    rels := [];
    for i in [1..n] do
        rels[i] := [];
        for j in [1..i-1] do
            w := Flat(WordOfVectorCR(ExponentsByPcp(g, g[i]^g[j])));
            if w <> [i,1] then rels[i][j] := w; fi;
        od;
        if r[i] > 0 then 
            w:= Flat(WordOfVectorCR(ExponentsByPcp(g, g[i]^r[i])));
            if w <> [] then rels[i][i] := w; fi;
        fi;
    od;

    return [r, rels];
end;

#############################################################################
##
#A PcpGroupCode(c)
##
PcpGroupCode := function(c)
    local r, n, coll, i, j;

    # set up collector
    coll := FromTheLeftCollector(Length(c[1]));

    # fill in relations
    for i in [1..Length(c[1])] do
        SetRelativeOrder(coll, i, c[1][i]);
        for j in [1..i-1] do
            if IsBound(c[2][i][j]) then
                SetConjugate(coll, i, j, c[2][i][j]);
            fi;
        od;
        if c[1][i] > 0 and IsBound(c[2][i][i]) then 
            SetPower(coll, i, c[2][i][i]);
        fi;
    od;

    UpdatePolycyclicCollector(coll);
    return PcpGroupByCollectorNC(coll);
end;

#############################################################################
##
#A SplitExtensionByPcpAndMats(G, mats, q)
##
SplitExtensionByPcpAndMats := function(G, mats, q)
    local g, n, m, rels, coll, i, j, w, c;

    g := Pcp(G);
    n := Length(g);
    m := Length(mats[1]);

    rels := RelativeOrdersOfPcp(g);
    coll := FromTheLeftCollector( n + m );

    for i in [1..n] do
        SetRelativeOrder(coll, i, rels[i]);
        for j in [1..i-1] do
            w := Flat(WordOfVectorCR(ExponentsByPcp(g, g[i]^g[j])));
            SetConjugate(coll, i, j, w);
        od;
        if rels[i]>0 then
            w := Flat(WordOfVectorCR(ExponentsByPcp(g, g[i]^rels[i])));
            SetPower(coll, i, w);
        fi;
    od;

    if q > 0 then
        for i in [n+1..n+m] do
            SetRelativeOrder(coll, i, q);
        od;
    fi;

    c := List([1..n], x -> 0);
    for i in [1..n] do
        for j in [n+1..n+m] do
            w := mats[i][j-n];
            if q > 0 then w := List(w, x -> x mod q); fi;
            w := Flat(WordOfVectorCR(Concatenation(c,w))); 
            SetConjugate(coll, j, i, w);
        od;
    od;

    UpdatePolycyclicCollector( coll );
    return PcpGroupByCollectorNC( coll );
end;

#############################################################################
##
#A WreathSpaceGroup(p, s, t)
##
## W_t(s) = C_t split P_{s-t}
##
WreathSpaceGroup := function(p,s,t)
    local q, r, Q, P, iso, g, I, W, G, B, i, l;

    # a brief check
    if not IsPrime(p) or s < t or s < 1 then return fail; fi;

    # set up
    q := p^(s-t);
    r := p^t;

    # construct point group
    if s = t then
        g := [];
        P := AbelianPcpGroup(1, [r]);
    elif s > t then 
        iso := IsomorphismPcpGroup(SylowSubgroup(SymmetricGroup(q),p));
        Q := Image(iso);
        g := List(Pcp(Q), x -> PreImagesRepresentative(iso,x));
        g := List(g, x -> PermutationMat(x,q));
        P := SplitExtensionByPcpAndMats( Q, g, r );
    fi;

    # construction action on translations
    I := IdentityMat(p^(t-1)*(p-1));
    W := CompanionMat(CyclotomicPolynomial(Rationals, r));
    G := List(g, x -> KroneckerProduct(x,I));
    B := [];
    for i in [1..q] do
        l := List([1..q], x -> I); l[i] := W;
        B[i] := DiagonalBlockMat(l);
    od;

    # the space group is T split P
    G :=  SplitExtensionByPcpAndMats(P, Concatenation(G,B), 0);
    SetHasFiniteCoclass(G, true);
    SetIsSpaceGroup(G, true);
    return G;
end;

#############################################################################
##
#A SpaceGroupByCoclass(p, r)
##
## Integral Space groups only!
##
NumberSpaceGroupsByCoclass := function(p,r)

    # check input 
    if not IsPrime(p) or not IsInt(r) or r < 1 then return fail; fi;

    # the maximal class case
    if r = 1 then return 1; fi;

    # the other cases
    if p = 2 then 
        if r = 2 then return 2; fi;
        if r = 3 then return 21; fi;
        if r = 4 then Print("this needs to be done\n"); fi;
    elif p = 3 then 
        if r = 2 then return 10; fi;
        if r = 3 then return 1271; fi;
        if r = 4 then return 137299952383; fi;
    elif p = 5 then
        if r = 2 then return 95; fi;
        if r = 3 then return 1110136753555665; fi;
    elif p = 7 then
        if r = 2 then return 4575; fi;
    elif p = 11 then
        if r = 2 then return 311076229; fi;
    fi;

    return fail;
end;

SpaceGroupByCoclass := function(p,r,i)
    local n, G;

    # a first check
    n := NumberSpaceGroupsByCoclass(p,r);
    if n = fail or i > n or i < 1 then return fail; fi;

    # get group
    G := fail; 
    if r = 1 then 
        G :=  WreathSpaceGroup(p,r,r); 
    elif p = 2 and r <= 3 then 
        G := PcpGroupCode(PSGTab2[r][i]); 
    elif p = 3 and r <= 3 then 
        G := PcpGroupCode(PSGTab3[r][i]); 
    elif p = 5 and r <= 2 then 
        G := PcpGroupCode(PSGTab5[r][i]); 
    fi;
  
    # no group found?
    if G = fail then return fail; fi;
        
    # otherwise add info
    SetHasFiniteCoclass(G, true);
    SetIsSpaceGroup(G, true);
    SetCoclass(G, r);

    # that's it
    return G;
end;

