

FreeMetab := function(p)
    local l, c, i, j, k, r, m;

    # set up
    r := p^2+3;
    c := FromTheLeftCollector(r);

    # relative orders and powers
    SetRelativeOrder(c, 1, p);
    SetRelativeOrder(c, 2, p);
    SetPower(c, 1, [p+3,1]);
    SetPower(c, 2, [3,1]);

    # conjugate b^a
    SetConjugate(c, 2, 1, [2,1,2*p+3,-1]);

    # conjugates x_i^a 
    for i in [0..p-2] do SetConjugate(c, i+3, 1, [i+4, 1]); od;
    SetConjugate(c, p+2, 1, [3, 1]);

    # conjugates y_j^a
    for j in [0..p-1] do SetConjugate(c, p+j+3, 1, [p+j+3, 1]); od;

    # conjugates x_i^b
    for i in [0..p-1] do SetConjugate(c, i+3, 2, [i+3, 1]); od;

    # conjugate y_j^b
    for j in [0..p-2] do SetConjugate(c, p+j+3, 2, [p+3+j+1, 1]); od;
    SetConjugate(c, p+p-1+3, 2, [p+3, 1]);

    # conjugates s_{ij}^a with i<p-1
    for i in [1..p-2] do
        for j in [1..p-1] do
            k := (i-1)*(p-1)+j+2*p+2;
            l := i*(p-1)+j+2*p+2;
            m := j+2*p+2;
            SetConjugate(c, k, 1, [m, -1, l, 1]);
        od;
    od;

    # conjugates s_{p-1,j}^a
    for j in [1..p-1] do
        k := (p-2)*(p-1)+j+2*p+2;
        m := j+2*p+2;
        SetConjugate(c, k, 1, [p+3,-1, p+3+j, 1, m, -1]);
    od;

    # conjugates s_{ij}^b with j<p-1
    for i in [1..p-1] do
        for j in [1..p-2] do
            k := (i-1)*(p-1)+j+2*p+2;
            l := (i-1)*(p-1)+1+2*p+2;
            m := (i-1)*(p-1)+j+1+2*p+2;
            SetConjugate(c, k, 2, [l, -1, m, 1]);
        od;
    od;

    # conjugates s_{i,p-1}^b
    for i in [1..p-1] do
        k := (i-1)*(p-1) + p-1 + 2*p + 2;
        l := (i-1)*(p-1)+1+2*p+2;
        SetConjugate(c, k, 2, [3, 1, i+3, -1, l, -1]);
    od;

    # final checks
    UpdatePolycyclicCollector(c);
    IsConfluent(c);

    return PcpGroupByCollector(c);
end;

FreeMetab2 := function(p)
    local F, f, G, g, h, K, k, A, a, w, l, c, v, i;

    F := FreeGroup(2);
    f := GeneratorsOfGroup(F);
    G := DirectProduct(CyclicGroup(p), CyclicGroup(p));
    g := GeneratorsOfGroup(G);
    h := GroupHomomorphismByImages(F,G,f,g);
    K := Kernel(h);
    k := FreeGeneratorsOfGroup(K);
    A := AbelianPcpGroup(p^2+1);
    a := Igs(A);
    w := GroupHomomorphismByImages(K,A,k,a);

    l := p^2+3;
    c := FromTheLeftCollector(l);

    # relative orders
    SetRelativeOrder(c, 1, p);
    SetRelativeOrder(c, 2, p);

    # powers
    v := ExponentsByIgs(a,Image(w, f[1]^p));
    v := Concatenation([0,0],v);
    SetPower(c, 1, WordByExps(v));
    v := ExponentsByIgs(a,Image(w, f[2]^p));
    v := Concatenation([0,0],v);
    SetPower(c, 2, WordByExps(v));

    # conjugate b^a = b k
    v := ExponentsByIgs(a,Image(w, f[2]^-1 * f[2]^f[1]));
    v := Concatenation([0,1],v);
    SetConjugate(c, 2, 1, WordByExps(v));

    for i in [1..l-2] do
        v := ExponentsByIgs(a,Image(w, k[i]^f[1]));
        v := Concatenation([0,0],v);
        SetConjugate(c, i+2, 1, WordByExps(v));
        v := ExponentsByIgs(a,Image(w, k[i]^f[2]));
        v := Concatenation([0,0],v);
        SetConjugate(c, i+2, 2, WordByExps(v));
    od;

    return PcpGroupByCollector(c);
end;
        

