#############################################################################
##
##  cobreadth.gi                Package coclass                  Bettina Eick
##

#############################################################################
##
#F Cobreadth( G, l )
##
InstallMethod( Cobreadth, true, [IsPcpGroup and IsFinite], 0,
function(G)
    local H, c, b, n;
    H := PcpGroupToPcGroup(RefinedPcpGroup(G));
    c := ConjugacyClasses(H);
    b := List(c, x -> Length(Factors(Size(x))));
    n := Length(Factors(Size(G)));
    return n - Maximum(b);
end );

#############################################################################
##
#F CobreadthRandom( G, b )
##
InstallMethod( CobreadthRandom, true, [IsPcpGroup and IsFinite], 0,
function(G)
    local m, g, C, c, r, l;
    m := infinity;
    l := 0;
    r := [];
    while l < 20 do
        g := Random(G);
        C := CentralizerNilpotentPcpGroup(G,g);
        c := Length(Factors(Size(C)));
        if IsBound(r[c]) then
            r[c] := r[c]+1;
        else
            r[c] := 1;
        fi;
        l := Maximum(r);
        m := Minimum( m, c );
    od;
    return m;
end );

InstallOtherMethod( CobreadthRandom, true, [IsPcpGroup, IsInt], 0,
function(G, b)
    local m, g, C, c, r, l;
    m := infinity;
    l := 0;
    r := [];
    while l < 20 and m < b do
        g := Random(G);
        C := CentralizerNilpotentPcpGroup(G,g);
        c := Length(Factors(Size(C)));
        if IsBound(r[c]) then
            r[c] := r[c]+1;
        else
            r[c] := 1;
        fi;
        l := Maximum(r);
        m := Minimum( m, c );
    od;
    return m;
end );

#############################################################################
##
#F CobreadthBound( G )
##
## An upper bound for cb(G) for infinite G.
##
InstallMethod( CobreadthBound, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    local T, nat, P, H, K, g, i, h, k, m, cl, b, c, r, e, d;

    # get translations
    T := Translations(G);
    nat := NaturalHomomorphism(G,T);

    # get point group
    P := Image(nat);
    H := RefinedPcpGroup(P);
    K := PcpGroupToPcGroup(H);

    # get gens and mats
    g := [];
    for i in [1..Length(Pcgs(K))] do
        h := Pcp(H)[i];                       
        k := PreImage(H!.bijection, h);      
        Add(g, PreImagesRepresentative(nat, k)); 
    od;
    m := LinearActionOnPcp(g, Pcp(T));

    # compute classes
    cl := ConjugacyClasses(K);

    # loop over classes
    b := infinity;
    for c in cl do
        g := Representative(c);
        r := Length(Factors(Size(StabilizerOfExternalSet(c))));
        if g <> g^0 then
            e := ExponentsOfPcElement(Pcgs(K), g);
            h := MappedVector(e, m);
            d := Determinant(h-h^0);
            if d <> 0 then
                d := Length(Factors(d));
                b := Minimum( b, d+r );
            fi;
        fi;
    od;
    return b;
end);

