
ClassPGroup := function(G)
    return Length(LowerCentralSeries(G))-1;
end;

CoclassPGroup := function(G)
    return Length(Factors(Size(G)))-ClassPGroup(G);
end;

#############################################################################
##
#F PrintScale
##
PrintScaleSheetClassTwo := function( Sheet, p, l, s )
    Text( Sheet, FONTS.normal, 20, 30, "|G|");
    Text( Sheet, FONTS.normal, 90, 30, "tree");
    Text( Sheet, FONTS.normal, 20, 70, 
          Concatenation( String(p), "^", String(l)));
end;

#############################################################################
##
#F StPresPcpGroup(G)
##
StPresPcpGroup := function( G )
    local p, c, H, P;
    p := Factors(Size(G))[1];
    c := ClassPGroup(G);
    H := PcpGroupToFpGroup(G);
    P := PqStandardPresentation( H : Prime:=p, ClassBound:=c);
    return ListPresentation(P);
end;

#############################################################################
##
#F IsomtypesFactors( G, sub, r )
##
IsomtypesFactors := function( G, sub, r )
    local fac, prs, i, j, c, n, p, P, F;

    # get factors and set up
    fac := List( sub, x -> G/x );
    prs := [];
    
    # get presentations
    for i in [1..Length(fac)] do
        c := ClassPGroup(fac[i]);
        n := Length(Factors(Size(fac[i])));
        p := G!.prime;
        if n-c <> r then 
            sub[i] := false;
            fac[i] := false;
            prs[i] := false;
        else
            F := PcpGroupToFpGroup(fac[i]);
            P := PqStandardPresentation( F : Prime:=p, ClassBound:=c);
            prs[i] := ListPresentation(P);
        fi;
    od;

    # discard groups
    sub := Filtered( sub, x -> x <> false );
    fac := Filtered( fac, x -> x <> false );
    prs := Filtered( prs, x -> x <> false );

    # determine isomorphisms
    for i in [1..Length(fac)] do
        if fac[i] <> false then
            j := Position(prs, prs[i]);
            if j = i then
                sub[i]!.isom := true;
            else
                sub[i]!.isom := j; 
            fi;
        fi;
    od;

    return sub;
end;

#############################################################################
##
#F DrawClassTwoBranch( G, l, reduce )
##
DrawClassTwoBranch := function( G, l, reduce )
    local p, s, d, n, t, a, T, A, U, V, W, pcp, grp, new, i, j, k,  
          cl, CR, tmp, com, g, v, title, Sheet, c;

    # set up
    p := G!.prime;
    s := G!.param;
    d := p^(s-1)*(p-1);

    # get class two module
    n := Position(RelativeOrdersOfPcp(Pcp(G)), 0)-1;
    t := Igs(G){[n+1..Length(Igs(G))]};
    a := t{[d+1..Length(t)]};
    T := SubgroupByIgs(G, t);
    A := SubgroupByIgs(G, a);

    # jump to some point in T/T' 
    U := T;
    G!.isNormal := true;
    while Index(G,U) < p^l do
        U := CommutatorSubgroup(G,U);
        U := SubgroupByIgsAndIgs( G, Igs(U), a );
    od;

    # set up graphic sheet
    title := Concatenation( "Branch for |G| = ", String(p), "^", String(n));
    Sheet := GraphicSheet( title, 600, 700 );
    PrintScaleSheetClassTwo( Sheet, p, l, 25 );

    # set up for iteration
    grp := [[U,[100, 70]]];
    Disc( Sheet, 100, 70, 3 );

    # iterate over normal subgroup in T'
    V := A;
    i := 0;
    while Length(grp) > 0 do
        i := i + 1;
        j := 1;
        Print("layer ",i," with ",Length(grp)," groups to go \n");

        W := CommutatorSubgroup(G, V);
        pcp := Pcp(V,W);
        V := W;

        new := [];
        for cl in grp do
            CR := rec( group  := cl[1],
                       super  := Pcp(G, cl[1]),
                       factor := Pcp(cl[1], GroupOfPcp(pcp)),
                       normal := pcp );
            AddFieldCR(CR); AddRelatorsCR(CR); 
            AddOperationCR(CR); AddInversesCR(CR);

            # get G-invariant complements
            tmp := InvariantComplementsCR( CR );
            Print("    determined ",Length(tmp)," complements \n");

            # reduce
            if reduce then
                com := IsomtypesFactors( G, tmp, G!.cocla );
                com := Filtered(com, x -> IsBool(x!.isom));
                Print("   reduced to ",Length(com)," isom types \n");
            else 
                com := Filtered(tmp, x -> CoclassPGroup(G/x)=G!.cocla);
                Print("   reduced to ",Length(com)," by coclass \n");
            fi;

            # draw groups and move on
            for k in [1..Length(com)] do
                repeat
                    v := PrintVertex(Sheet,i,j,false,cl[2]);
                    j := j + 1;
                until not IsBool(v);
                Add( new, [com[k],v] );
            od;
        od;

        # next layer
        grp := new;
    od;
end;

