
OutGroupOrder := function(G)
    local A, B;
    IsPGroup(G);
    A := Size(AutomorphismGroup(G));
    B := Size(G)/Size(Center(G));
    return Collected(Factors(A/B));
end;

fold := function (p, a, x)
   local i, j;
   j := p;
   for i in [1..x] do
      j := j^(a^3) * j^-1;
   od;
   return j;
end;

# j: branch has root of class j 
# k: is the class of the group = j+n where n is depth
# c: is 0 or 1

# j in [9,12,15,...]
# k in [j..2*j], wobei 2*j-2 schon merkwuerdige Aut Group hat.

SkeletonGroupH0Fp := function( j, k, c )
    local x, n, F, f, a, b, T, t, R, S, lhs, rhs, prd, p, i;

    # check
    if not j >= 7 then return fail; fi;
    if not j <= k then return fail; fi;
    x := j/3 - 1; if not IsInt(x) then return fail; fi;

    # set up
    n := k-j;

    # new version
    F := FreeGroup (3 + k + x +1);
    f := GeneratorsOfGroup(F);
    a := f[1]; b := f[2];
    T := f{[3..k+3]};
    t := T[1];

    R := [ a^9, (t*a^3)^3, Comm(Comm(t,a^3),t),
           Comm(t,t^(a^4))*Comm(t^a,t^(a^3)),
           Comm(t,t^(a^5))*Comm(t^(a^2),t^(a^3)),
           b^3, a^b/a^4,T[k+1], T[1]^b/T[1] ];

    S := List([2..k+1], i -> T[i]/Comm(T[i - 1], a));
    R := Concatenation(R, S);

    lhs := Comm(t, t^a);
    prd := t^(a^-1) * (t^-1)^(a^2);
    Add(R, f[3 + k + 1]/prd);

    for i in [1..x] do 
       p := f[3 + k +i];
       p := p^(a^3) * p^-1;
       Add(R, f[3 + k+i + 1]/p);
    od;

    prd := f[3 + k + x + 1];
    rhs := prd * ((prd^3)^(a^3) * prd^-3)^c;
    Add(R, lhs/rhs);

    return F/R;

    # old version
    F := FreeGroup(k+3);
    f := GeneratorsOfGroup(F);
    a := f[1]; b := f[2]; 
    T := f{[3..k + 3]};
    t := T[1];

    R := [ a^9, (t*a^3)^3, Comm(Comm(t,a^3),t),
           Comm(t,t^(a^4))*Comm(t^a,t^(a^3)),
           Comm(t,t^(a^5))*Comm(t^(a^2),t^(a^3)),
           b^3, a^b/a^4,T[k+1], T[1]^b/T[1] ];

    S := List([2..k+1], i -> T[i]/Comm(T[i - 1], a));

    lhs := Comm(t, t^a);
    prd := t^(a^-1) * (t^-1)^(a^2);
    prd := fold (prd, a, x);
    rhs := prd * ((prd^3)^(a^3) * prd^-3)^c;

    R := Concatenation(S, R, [lhs/rhs]);

    return F/R;
end;

SkeletonGroupH0 := function( j, k, c )
    local Q;
    Q := SkeletonGroupH0Fp( j, k, c );
    return Pq( Q : Prime := 3, ClassBound := k );
end;

# j in [10,13,16,...]
# k in [j..2*j-2]

SkeletonGroupH1 := function(j, k, c)
    local Q, G, g, f, U, i;
    Q := SkeletonGroupH0Fp(j-1, k+1, c);
    i := PqEpimorphism( Q : Prime := 3, Classbound := k+1 );
    G := Image(i);
    g := List(GeneratorsOfGroup(Q){[1,2,4]}, x -> Image(i,x));
    U := Subgroup(G, g);
    return PcGroupWithPcgs(Pcgs(U));
end;

# j in [11,14,17,...]
# k in [j..2*j-4]

SkeletonGroupH2 := function(j, k, c)
    local G, g, Q, f, U, i;
    Q := SkeletonGroupH0Fp(j-2, k+2, c);
    i := PqEpimorphism( Q : Prime := 3, Classbound := k+2 );
    G := Image(i);
    g := List(GeneratorsOfGroup(Q){[1,2,5]}, x -> Image(i,x));
    U := Subgroup(G, g);
    return PcGroupWithPcgs(Pcgs(U));
end;

BranchLimitH0 := function(j)
    return j-9; 
end;

BranchLimitH1 := function(j)
    return j-11; 
end;

BranchLimitH2 := function(j)
    return j-13; 
end;

PrintTwigH0 := function(G, i,j,k)
    local a, b, file, ii, jj, kk;
    ii := String(i);
    jj := String(j);
    kk := String(k);
    file := Concatenation("twigsH/twig0_",ii,"_",jj,"_",kk,".gi");
    PrintTo(file,"TwigH0_",ii,"_",jj,"_",kk," := [\n");
    for a in [1..Length(G)] do
        AppendTo(file,"[");
        for b in [1..Length(G[a])] do
            AppendTo(file,"[",G[a][b][1],",",G[a][b][2],"]");
            if b < Length(G[a]) then
                AppendTo(file,",");
            else
                AppendTo(file,"]");
            fi;
        od;
        if a < Length(G) then
            AppendTo(file,",\n");
        else
            AppendTo(file,"]\n");
        fi;
    od;
    AppendTo(file,";\n\n");
    AppendTo(file,"TWIGH0[",ii,"][",jj,"][",kk,"]:= ");
    AppendTo(file,"TwigH0_",ii,"_",jj,"_",kk,";\n");
    AppendTo("summH.gi", "Add(listH0, [",ii,",",jj,",",kk,"]);\n");
end;

PrintTwigH1 := function(G, i,j,k)
    local a, b, file, ii, jj, kk;
    ii := String(i);
    jj := String(j);
    kk := String(k);
    file := Concatenation("twigsH/twig1_",ii,"_",jj,"_",kk,".gi");
    PrintTo(file,"TwigH1_",ii,"_",jj,"_",kk," := [\n");
    for a in [1..Length(G)] do
        AppendTo(file,"[");
        for b in [1..Length(G[a])] do
            AppendTo(file,"[",G[a][b][1],",",G[a][b][2],"]");
            if b < Length(G[a]) then
                AppendTo(file,",");
            else
                AppendTo(file,"]");
            fi;
        od;
        if a < Length(G) then
            AppendTo(file,",\n");
        else
            AppendTo(file,"]\n");
        fi;
    od;
    AppendTo(file,";\n\n");
    AppendTo(file,"TWIGH1[",ii,"][",jj,"][",kk,"]:= ");
    AppendTo(file,"TwigH1_",ii,"_",jj,"_",kk,";\n");
    AppendTo("summH.gi", "Add(listH1, [",ii,",",jj,",",kk,"]);\n");
end;

PrintTwigH2 := function(G, i,j,k)
    local a, b, file, ii, jj, kk;
    ii := String(i);
    jj := String(j);
    kk := String(k);
    file := Concatenation("twigsH/twig2_",ii,"_",jj,"_",kk,".gi");
    PrintTo(file,"TwigH2_",ii,"_",jj,"_",kk," := [\n");
    for a in [1..Length(G)] do
        AppendTo(file,"[");
        for b in [1..Length(G[a])] do
            AppendTo(file,"[",G[a][b][1],",",G[a][b][2],"]");
            if b < Length(G[a]) then
                AppendTo(file,",");
            else
                AppendTo(file,"]");
            fi;
        od;
        if a < Length(G) then
            AppendTo(file,",\n");
        else
            AppendTo(file,"]\n");
        fi;
    od;
    AppendTo(file,";\n\n");
    AppendTo(file,"TWIGH2[",ii,"][",jj,"][",kk,"]:= ");
    AppendTo(file,"TwigH2_",ii,"_",jj,"_",kk,";\n");
    AppendTo("summH.gi", "Add(listH2, [",ii,",",jj,",",kk,"]);\n");
end;

GetTwigInH0 := function(i, j, k)
    local l, G, bran, T, S;
    l := k+9;
    if k < BranchLimitH0(j) then 
        G := SkeletonGroupH0( j, j+l+1, i-1 );
        T := CoclassTree(G, 3, j+l, true);
    else
        G := SkeletonGroupH0( j, j+l, i-1 );
        T := CoclassTree(G, 3, j+l, false);
    fi;
    S := TwigCanoForm(T);
    PrintTwigH0(S, i, j, k);
end;

GetTwigInH1 := function(i, j, k)
    local l, G, bran, T, S;
    l := k+9;
    if k < BranchLimitH1(j) then
        G := SkeletonGroupH1( j, j+l+1, i-1 );
        T := CoclassTree(G, 3, j+l, true);
    else
        G := SkeletonGroupH1( j, j+l, i-1 );
        T := CoclassTree(G, 3, j+l, false);
    fi;
    S := TwigCanoForm(T);
    PrintTwigH1(S, i, j, k);
end;

GetTwigInH2 := function(i, j, k)
    local l, G, bran, T, S;
    l := k+9;
    if k < BranchLimitH2(j) then
        G := SkeletonGroupH2( j, j+l+1, i-1 );
        T := CoclassTree(G, 3, j+l, true);
    else
        G := SkeletonGroupH2( j, j+l, i-1 );
        T := CoclassTree(G, 3, j+l, false);
    fi;
    S := TwigCanoForm(T);
    PrintTwigH2(S, i, j, k);
end;

AllSkeletonGroupsH0 := function( j, k )
    local grps;
    grps := List([0,1], c -> SkeletonGroupH0(j,k,c));
    return List(grps, FpGroupPcGroup);
end;

AllSkeletonGroupsH1 := function( j, k )
    local grps;
    grps := List([0,1], c -> SkeletonGroupH1(j,k,c));
    return List(grps, FpGroupPcGroup);
end;

AllSkeletonGroupsH2 := function( j, k )
    local grps;
    grps := List([0,1], c -> SkeletonGroupH2(j,k,c));
    return List(grps, FpGroupPcGroup);
end;

