
DescOfGroup := function(G, a, b)
    local D, t, c, s, e, d;
    D := [ [[false, G[a][b][2]]] ]; 
    t := [b];
    for c in [a+1..Length(G)] do
        s := [];  
        D[c-a+1] := [];
        for d in [1..Length(G[c])] do
            e := Position(t, G[c][d][1]);
            if not IsBool(e) then 
                Add(D[c-a+1], [e, G[c][d][2]]);
                Add(s, d);
            fi;
        od;
        if Length(D[c-a+1]) = 0 then 
            Unbind(D[c-a+1]);
            return D;
        fi;
        t := s;
    od;
    return D; 
end;

MaxSubsOfTree := function(G)
    return List([1..Length(G[2])], x -> DescOfGroup(G,2,x));
end;

AddMaxSub := function(G, H)
    local r, i, s, j;

    # the first layer
    if not IsBound(G[2]) then 
        G[2] := [ [1, H[1][1][2]] ]; 
    else
        Add(G[2], [1, H[1][1][2]] );
    fi;
    r := [Length(G[2])];

    # the other layers
    for i in [2..Length(H)] do
        if not IsBound(G[i+1]) then G[i+1] := []; fi;
        s := [];
        for j in [1..Length(H[i])] do
            Add(G[i+1], [r[H[i][j][1]],H[i][j][2]]);
            Add(s, Length(G[i+1]));
        od;
        r := s;
    od;
end;

TreeByMaxSubs := function(mul, list)
    local G, i;
    G := [[[false,mul]]];
    for i in [1..Length(list)] do
        AddMaxSub(G, list[i]);
    od;
    return G;
end;

SubTreeIsom := function(S, T)
    return S{[2..Length(S)]} = T{[2..Length(T)]};
end;

FullTreeIsom := function(S, T)
    return S = T;
end;

TwigLess := function(S, T)
    local d1, d2, i;
    if FullTreeIsom(S,T) then return true; fi;
    if S[1][1][2] < T[1][1][2] then return true; fi;
    if S[1][1][2] > T[1][1][2] then return false; fi;
    if Length(S[2])<Length(T[2]) then return true; fi;
    if Length(S[2])>Length(T[2]) then return false; fi;
    d1 := MaxSubsOfTree(S);
    d2 := MaxSubsOfTree(T);
    for i in [1..Length(d1)] do
        if not FullTreeIsom(d1[i],d2[i]) then
            return TwigLess(d1[i],d2[i]);
        fi;
    od;
    Error("should not happen");
end;

CollectTwigs := function( cano )
    local i, j;
    for i in [2..Length(cano)] do
        j := 1;
        while j < i and not IsBool(cano[i]) do 
            if not IsBool(cano[j]) and SubTreeIsom(cano[j],cano[i]) then 
                cano[j][1][1][2] := cano[j][1][1][2]+cano[i][1][1][2];
                cano[i] := false;
            fi;
            j := j+1;
        od;
    od;
    return Filtered(cano, x -> not IsBool(x));
end;

TwigCanoForm := function(G)
    local desc, cano;

    # check
    if Length(G) = 1 then return G; fi;
 
    # get subtrees at level 2
    desc := MaxSubsOfTree(G);

    # compute their cano form
    cano := List(desc, TwigCanoForm);

    # collect
    cano := CollectTwigs(cano);

    # sort
    Sort(cano, TwigLess);

    # reconstruct tree
    return TreeByMaxSubs( G[1][1][2], cano);
end;

DepthOfTwig := function(G)
    return Length(G)-1;
end;

GroupsAtLevel := function(G,l)
    local s, i, t, k, j; 
    s := 0;
    for i in [1..Length(G[l])] do
        t := G[l][i][2];
        k := G[l][i][1];
        for j in Reversed([1..l-1]) do
            t := t*G[j][k][2];
            k := G[j][k][1];
        od;
        s := s + t; 
    od;
    return s;
end;

GroupsAtLevels := function( G )
    return List([1..Length(G)], x -> GroupsAtLevel(G,x));
end;

WidthOfTwig := function(G)
    return Maximum(GroupsAtLevels(G));
end;

SizeOfTwig := function(G);
    return Sum(GroupsAtLevels(G));
end;

