#############################################################################
##
#W tree.gi                 Package Coclass                      Bettina Eick
##

Read("/Users/beick/Home/gap/gap4r8/local/pkg/coclass/gap/trees/tcano.gi");

#############################################################################
##
#F Some helper functions
##

IsTerminal := function( G ) return not IsCapable(G); end;

EliminateMainLine := function( K, cap, aut )
    local sub;
    sub := Filtered([1..Length(cap)], x -> aut[x] = aut[1]);
    while Length(sub) > 1 do
        if IsIsomorphicPGroup( K, cap[sub[1]] ) then
            sub := [sub[1]];
        else
            sub := sub{[2..Length(sub)]};
        fi;
    od;
    sub := Difference( [1..Length(cap)], sub );
    return cap{sub};
end;

CoverInfo := function(G)
    local s, P, l, d;
    s := Length(PCentralSeries(G));
    P := PqPCover(G);
    l := PCentralSeries(P)[s];
    d := Length(Factors(Size(P)/Size(G)));
    return [d, Length(Factors(Size(l)))];
end;

GaloisSize := function(G)
    local p, a;
    if IsBound(G!.gal) then return G!.gal; fi;
    p := PrimePGroup(G);
    a := Size(AutomorphismGroup(G));
    G!.gal := Product(Filtered(Factors(a), x -> x <> p));
    return G!.gal;
end;

#############################################################################
##
#F CoclassTree(G, p, s, bran)
##
CoclassTree := function( G, p, s, bran)
    local l, F, H, K, tree, i, j, d, des, ter, cap, aut;

    # set up p-quotient of G
    if IsPcpGroup(G) then 
        F := PcpGroupToFpGroup(G);
    elif IsPcGroup(G) then 
        F := FpGroupPcGroup(G);
    elif IsFpGroup(G) then 
        F := G;
    fi;
    H := Pq(F : Prime := p, ClassBound := s);

    # if desired then determine next class
    if bran then 
        K := Pq(F : Prime := p, ClassBound := s+1); 
    fi;

    # construct tree
    tree := [[[false, H]]]; i := 1;
    while Length(tree[i]) > 0 do
        i := i+1;
        tree[i] := [];
        for j in [1..Length(tree[i-1])] do
            Print("start group ",j," at level ",i-1,"\n");
            d := tree[i-1][j][2];
            if IsGroup(d) then 
                des := PqDescendants( d : StepSize := 1 );
                ter := Filtered( des, IsTerminal );
                cap := Filtered( des, IsCapable );
            
                # cut of main line if desired
                if bran and i = 2 then 
                    aut := List( cap, x -> Size(AutomorphismGroup(x)) );
                    SortParallel( aut, cap, function(a,b) return a>b; end );
                    cap := EliminateMainLine( K, cap, aut ); 
                fi;

                # add capables
                Append( tree[i], List(cap, g -> [j,g]) );

                # add terminals
                Add( tree[i], [j, Length(ter)] );

                # get rid of original group
                tree[i-1][j][2] := 1;
            fi;
        od;
    od;
    Unbind(tree[i]);

    return tree;
end;


#############################################################################
##
#F CoclassBranch 
##
## Determines the descendant tree of the k-th layer 1 group in branch s.
## Computes capable groups only.
##
CoclassBranch := function( G, p, s, k )
    local F, H, K, d, a, tree, i, j, cap, ter;

    # set up p-quotient of G
    if IsPcpGroup(G) then 
        F := PcpGroupToFpGroup(G);
    elif IsPcGroup(G) then 
        F := FpGroupPcGroup(G);
    elif IsFpGroup(G) then 
        F := G;
    fi;
    H := Pq(F : Prime := p, ClassBound := s);
    K := Pq(F : Prime := p, ClassBound := s+1); 

    # get first level descendants
    d := PqDescendants( H : StepSize := 1, AllDescendants := false );
    a := List( d, x -> Size(AutomorphismGroup(x)) );
    SortParallel( a, d, function(a,b) return a>b; end );

    # eliminate the main line
    d := EliminateMainLine( K, d, a );
    if Length(d) <> (p-3)/2 then Error("something wrong here"); fi;

    # choose group k
    H := d[k];

    # construct tree
    tree := [[[false, H]]]; i := 1;
    while Length(tree[i]) > 0 and i <= s do
        i := i+1;
        tree[i] := [];
        for j in [1..Length(tree[i-1])] do
            d := tree[i-1][j];
            if IsGroup(d[2]) then 

            Print("start group ",j," at level ",i-1, " parent ", d[1],"\n");

            # get descendants
            cap := PqDescendants( d[2] : StepSize:=1, AllDescendants:=false );

            # filter
            ter := Filtered(cap, x -> GaloisSize(x)=1);
            cap := Filtered(cap, x -> GaloisSize(x)=6);

            # add info
            cap := List(cap, x -> [j,x]);
 
            # add to tree
            Append( tree[i], cap );
            if Length(ter) > 0 then Add( tree[i], [j, Length(ter)]); fi;

            # reduce
            tree[i-1][j][2] := 1;

            fi;
        od;
    od;
    Unbind(tree[i]);

    tree := TwigCanoForm(tree);

    for i in [1..Length(tree)] do Print(tree[i],"\n"); od;
end;

CoclassGaloisTwig := function( G, p, s, k, t )
    local F, H, K, d, a, tree, i, j, cap, ter, h;

    # set up p-quotient of G
    if IsPcpGroup(G) then 
        F := PcpGroupToFpGroup(G);
    elif IsPcGroup(G) then 
        F := FpGroupPcGroup(G);
    elif IsFpGroup(G) then 
        F := G;
    fi;
    H := Pq(F : Prime := p, ClassBound := s);
    K := Pq(F : Prime := p, ClassBound := s+1); 

    # get first level descendants
    d := PqDescendants( H : StepSize := 1, AllDescendants := false );
    a := List( d, x -> Size(AutomorphismGroup(x)) );
    SortParallel( a, d, function(a,b) return a>b; end );

    # eliminate the main line
    d := EliminateMainLine( K, d, a );
    if Length(d) <> (p-3)/2 then Error("something wrong here"); fi;

    # choose group k
    H := d[k];

    # go down to desired descendant
    h := 1;
    while GaloisSize(H) <> t do 
        cap := PqDescendants( H : StepSize:=1, AllDescendants:=false );
        H := First(cap, x -> GaloisSize(x)=t);
        if H = fail then H := First(cap, x -> GaloisSize(x)=6); fi;
        h := h+1;
    od;
       
    # construct tree
    tree := [[[false, H]]]; i := 1;
    while Length(tree[i]) > 0 and i+h <= s do
        i := i+1;
        tree[i] := [];
        for j in [1..Length(tree[i-1])] do
            d := tree[i-1][j];
            if IsGroup(d[2]) then 

            Print("start group ",j," at level ",i-1, " parent ", d[1],"\n");

            # get descendants
            cap := PqDescendants( d[2] : StepSize:=1, AllDescendants:=false );

            # filter
            cap := Filtered(cap, x -> GaloisSize(x)=t);

            # add info
            cap := List(cap, x -> [j,x]);
 
            # add to tree
            Append( tree[i], cap );

            # reduce
            tree[i-1][j][2] := 1;

            fi;
        od;
    od;
    Unbind(tree[i]);

    tree := TwigCanoForm(tree);

    for i in [1..Length(tree)] do Print(tree[i],"\n"); od;
end;


